from __future__ import annotations

from .anvil_1484 import Anvil1484Interface as ParentInterface


class Anvil1503Interface(ParentInterface):
    """
    Changed height keys again
    """

    def __init__(self):
        super().__init__()
        self._set_feature("height_map", "C|V3")

    @staticmethod
    def minor_is_valid(key: int):
        return 1503 <= key < 1519


export = Anvil1503Interface



from __future__ import annotations
from abc import abstractmethod, ABC

from typing import (
    List,
    Tuple,
    Iterable,
    TYPE_CHECKING,
    Any,
    Dict,
    Callable,
    Sequence,
    Union,
    Type,
)
import numpy


from amulet_nbt import (
    AbstractBaseTag,
    IntTag,
    ListTag,
    CompoundTag,
    NamedTag,
    AnyNBT,
)

from amulet.api.chunk import Chunk, StatusFormats
from amulet.api.wrapper import Interface
from amulet.level import loader
from amulet.api.data_types import AnyNDArray, VersionIdentifierType
from amulet.api.wrapper import EntityIDType, EntityCoordType

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator
    from amulet.api.block_entity import BlockEntity
    from amulet.api.entity import Entity


ChunkDataType = Dict[str, NamedTag]

ChunkPathType = Tuple[
    str,  # The layer name
    Sequence[Tuple[Union[str, int], Type[AbstractBaseTag]]],
    Type[AbstractBaseTag],
]


class BaseDecoderEncoder(ABC):
    def __init__(self):
        self.__decoders = {}
        self.__post_decoders = {}
        self.__encoders = {}
        self.__post_encoders = {}

    def _register_decoder(self, decoder: Callable):
        """Register a function that does the decoding"""
        self.__decoders[decoder] = None

    def _register_post_decoder(self, post_decoder: Callable):
        """Register a function that runs after the decoding"""
        self.__post_decoders[post_decoder] = None

    def _unregister_decoder(self, decoder: Callable):
        """Unregister a function that does the decoding"""
        del self.__decoders[decoder]

    def _unregister_post_decoder(self, post_decoder: Callable):
        """Unregister a function that runs after the decoding"""
        del self.__post_decoders[post_decoder]

    def _do_decode(self, *args, **kwargs):
        for decoder in self.__decoders:
            decoder(*args, **kwargs)
        for post_decoder in self.__post_decoders:
            post_decoder(*args, **kwargs)

    def _register_encoder(self, encoder: Callable):
        """Register a function that does the encoding"""
        self.__encoders[encoder] = None

    def _register_post_encoder(self, post_encoder: Callable):
        """Register a function that runs after the encoding"""
        self.__post_encoders[post_encoder] = None

    def _unregister_encoder(self, encoder: Callable):
        """Unregister a function that does the encoding"""
        del self.__encoders[encoder]

    def _unregister_post_encoder(self, post_encoder: Callable):
        """Unregister a function that runs after the encoding"""
        del self.__post_encoders[post_encoder]

    def _do_encode(self, *args, **kwargs):
        for encoder in self.__encoders:
            encoder(*args, **kwargs)
        for post_encoder in self.__post_encoders:
            post_encoder(*args, **kwargs)


class BaseAnvilInterface(Interface, BaseDecoderEncoder):
    # The chunk object, the chunk data, the floor chunk coord, the chunk height (in sub-chunks)
    DecoderType = Callable[[Chunk, ChunkDataType, int, int], None]
    EncoderType = Callable[[Chunk, ChunkDataType, int, int], None]
    _register_decoder: Callable[[DecoderType], None]
    _register_post_decoder: Callable[[DecoderType], None]
    _unregister_decoder: Callable[[DecoderType], None]
    _unregister_post_decoder: Callable[[DecoderType], None]
    _do_decode: DecoderType
    _register_encoder: Callable[[EncoderType], None]
    _register_post_encoder: Callable[[EncoderType], None]
    _unregister_encoder: Callable[[EncoderType], None]
    _unregister_post_encoder: Callable[[EncoderType], None]
    _do_encode: EncoderType

    def __init__(self):
        BaseDecoderEncoder.__init__(self)
        self._feature_options = {
            "status": StatusFormats,
            "height_map": [
                "256IARequired",  # A 256 element Int Array in HeightMap
                "256IA",  # A 256 element Int Array in HeightMap
                "C|V1",  # A Compound of Long Arrays with these keys "LIQUID", "SOLID", "LIGHT", "RAIN"
                "C|V2",  # A Compound of Long Arrays with these keys "WORLD_SURFACE_WG", "OCEAN_FLOOR_WG", "MOTION_BLOCKING", "MOTION_BLOCKING_NO_LEAVES", "OCEAN_FLOOR", "LIGHT_BLOCKING"
                "C|V3",  # A Compound of Long Arrays with these keys "WORLD_SURFACE_WG", "OCEAN_FLOOR_WG", "MOTION_BLOCKING", "MOTION_BLOCKING_NO_LEAVES", "OCEAN_FLOOR", "LIGHT_BLOCKING", "WORLD_SURFACE"
                "C|V4",  # A Compound of Long Arrays with these keys "WORLD_SURFACE_WG", "OCEAN_FLOOR_WG", "MOTION_BLOCKING", "MOTION_BLOCKING_NO_LEAVES", "OCEAN_FLOOR", "WORLD_SURFACE"
            ],
            # 'carving_masks': ['C|?BA'],
            "light_optional": ["false", "true"],
            "block_entity_format": [EntityIDType.namespace_str_id],
            "block_entity_coord_format": [EntityCoordType.xyz_int],
            "entity_format": [EntityIDType.namespace_str_id],
            "entity_coord_format": [EntityCoordType.Pos_list_double],
            # 'lights': [],
        }
        self._features = {key: None for key in self._feature_options.keys()}

    def _set_feature(self, feature: str, option: Any):
        assert feature in self._feature_options, f"{feature} is not a valid feature."
        assert (
            option is None or option in self._feature_options[feature]
        ), f'Invalid option {option} for feature "{feature}"'
        self._features[feature] = option

    def is_valid(self, key: Tuple) -> bool:
        return key[0] == "java" and self.minor_is_valid(key[1])

    @staticmethod
    @abstractmethod
    def minor_is_valid(key: int):
        raise NotImplementedError

    def get_translator(
        self,
        max_world_version: VersionIdentifierType,
        data: ChunkDataType = None,
    ) -> Tuple["Translator", int]:
        if data is None:
            key = max_world_version
            version = max_world_version[1]
        else:
            data_version = (
                data.get("region", NamedTag(CompoundTag()))
                .compound.get("DataVersion", IntTag(-1))
                .py_int
            )
            key, version = (("java", data_version), data_version)

        return loader.Translators.get(key), version

    def get_layer_obj(
        self,
        obj: ChunkDataType,
        data: Tuple[
            str,
            Sequence[Tuple[Union[str, int], Type[AbstractBaseTag]]],
            Union[None, AnyNBT, Type[AbstractBaseTag]],
        ],
        *,
        pop_last=False,
    ) -> Any:
        """
        Get an object from a nested NBT structure layer

        :param obj: The chunk data object
        :param data: The data layer name, the nbt path and the default
        :param pop_last: If true the last key will be popped
        :return: The found data or the default
        """
        layer_key, path, default = data
        if layer_key in obj:
            return self.get_nested_obj(
                obj[layer_key].compound, path, default, pop_last=pop_last
            )
        elif default is None or isinstance(default, AbstractBaseTag):
            return default
        elif issubclass(default, AbstractBaseTag):
            return default()
        else:
            raise TypeError("default must be None, an NBT instance or an NBT class.")

    def set_layer_obj(
        self,
        obj: ChunkDataType,
        data: Tuple[
            str,
            Sequence[Tuple[Union[str, int], Type[AbstractBaseTag]]],
            Union[None, AnyNBT, Type[AbstractBaseTag]],
        ],
        default_tag: AnyNBT = None,
        *,
        setdefault=False,
    ) -> AnyNBT:
        """
        Setdefault on a ChunkDataType object

        :param obj: The ChunkDataType object to use
        :param data: The data to set
        :param setdefault: If True will behave like setdefault. If False will replace existing data.
        :return: The existing data found or the default that was set
        """
        layer_key, path, default = data
        default = default if default_tag is None else default_tag
        if not path:
            raise ValueError("was not given a path to set")
        tag = obj.setdefault(layer_key, NamedTag()).compound
        *path, (key, dtype) = path
        if path:
            key_path = next(zip(*path))
        else:
            key_path = ()
        return self.set_obj(
            tag, key, dtype, default, path=key_path, setdefault=setdefault
        )

    @abstractmethod
    def decode(
        self, cx: int, cz: int, data: ChunkDataType, bounds: Tuple[int, int]
    ) -> Tuple["Chunk", AnyNDArray]:
        """
        Create an amulet.api.chunk.Chunk object from raw data.
        :param cx: chunk x coordinate
        :param cz: chunk z coordinate
        :param data: The chunk data
        :param bounds: The minimum and maximum bounds of the chunk. In 1.17 this is required to define where the biome array sits.
        :return: Chunk object in version-specific format, along with the block_palette for that chunk.
        """
        raise NotImplementedError

    def _decode_entity_list(self, entities: ListTag) -> List["Entity"]:
        entities_out = []
        if entities.list_data_type == CompoundTag.tag_id:
            for nbt in entities:
                entity = self._decode_entity(
                    NamedTag(nbt),
                    self._features["entity_format"],
                    self._features["entity_coord_format"],
                )
                if entity is not None:
                    entities_out.append(entity)

        return entities_out

    def _decode_block_entity_list(self, block_entities: ListTag) -> List["BlockEntity"]:
        entities_out = []
        if block_entities.list_data_type == CompoundTag.tag_id:
            for nbt in block_entities:
                if not isinstance(nbt, CompoundTag):
                    continue
                entity = self._decode_block_entity(
                    NamedTag(nbt),
                    self._features["block_entity_format"],
                    self._features["block_entity_coord_format"],
                )
                if entity is not None:
                    entities_out.append(entity)

        return entities_out

    @abstractmethod
    def encode(
        self,
        chunk: "Chunk",
        palette: AnyNDArray,
        max_world_version: Tuple[str, int],
        bounds: Tuple[int, int],
    ) -> ChunkDataType:
        """
        Encode a version-specific chunk to raw data for the format to store.

        :param chunk: The already translated version-specific chunk to encode.
        :param palette: The block_palette the ids in the chunk correspond to.
        :type palette: numpy.ndarray[Block]
        :param max_world_version: The key to use to find the encoder.
        :param bounds: The minimum and maximum bounds of the chunk. In 1.17 this is required to define where the biome array sits.
        :return: Raw data to be stored by the Format.
        """
        raise NotImplementedError

    def _encode_entity_list(self, entities: Iterable["Entity"]) -> ListTag:
        entities_out = []
        for entity in entities:
            nbt = self._encode_entity(
                entity,
                self._features["entity_format"],
                self._features["entity_coord_format"],
            )
            if nbt is not None:
                entities_out.append(nbt.compound)

        return ListTag(entities_out)

    def _encode_block_entity_list(
        self, block_entities: Iterable["BlockEntity"]
    ) -> ListTag:
        entities_out = []
        for entity in block_entities:
            nbt = self._encode_block_entity(
                entity,
                self._features["block_entity_format"],
                self._features["block_entity_coord_format"],
            )
            if nbt is not None:
                entities_out.append(nbt.compound)

        return ListTag(entities_out)



from __future__ import annotations
import copy
import numpy

from amulet_nbt import NamedTag
from amulet.api.data_types import BlockCoordinates
from .abstract_base_entity import AbstractBaseEntity


class BlockEntity(AbstractBaseEntity):
    """
    A class to contain all the data to define a BlockEntity.
    """

    obj_name = "BlockEntity"
    coord_types = (int, numpy.integer)

    def __init__(
        self,
        namespace: str,
        base_name: str,
        x: int,
        y: int,
        z: int,
        nbt: NamedTag,
    ):
        """
        Constructs a :class:`BlockEntity` instance.

        :param namespace: The namespace of the block entity eg "minecraft"
        :param base_name: The base name of the block entity eg "chest"
        :param x: The x coordinate of the block entity
        :param y: The y coordinate of the block entity
        :param z: The z coordinate of the block entity
        :param nbt: The NBT stored with the block entity
        """
        super().__init__(namespace, base_name, x, y, z, nbt)

    @property
    def x(self) -> int:
        """The x location of the BlockEntity. Read Only"""
        return self._x

    @property
    def y(self) -> int:
        """The y location of the BlockEntity. Read Only"""
        return self._y

    @property
    def z(self) -> int:
        """The z location of the BlockEntity. Read Only"""
        return self._z

    @property
    def location(self) -> BlockCoordinates:
        """The location of the BlockEntity. Read Only"""
        return self._x, self._y, self._z

    def new_at_location(self, x: int, y: int, z: int) -> BlockEntity:
        """
        Creates a copy of this BlockEntity at a new location
        BlockEntities are stored in the chunk based on their location so location cannot be mutable
        """
        return BlockEntity(
            self._namespace,
            self._base_name,
            x,
            y,
            z,
            copy.deepcopy(self._nbt),
        )



from __future__ import annotations

from typing import List, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from amulet.api.chunk import Chunk

import amulet
from amulet.api.entity import Entity
from amulet_nbt import ListTag, IntTag

from .base_anvil_interface import ChunkDataType, ChunkPathType

from .anvil_2529 import Anvil2529Interface as ParentInterface


class Anvil2681Interface(ParentInterface):
    """
    Entities moved to a different storage layer
    """

    EntitiesDataVersion: ChunkPathType = (
        "entities",
        [("DataVersion", IntTag)],
        IntTag,
    )
    EntityLayer: ChunkPathType = (
        "entities",
        [("Entities", ListTag)],
        ListTag,
    )

    @staticmethod
    def minor_is_valid(key: int):
        return 2681 <= key < 2709

    def _post_decode_data_version(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        super()._post_decode_data_version(chunk, data, floor_cy, height_cy)
        self.get_layer_obj(data, self.EntitiesDataVersion, pop_last=True)

    def _decode_entities(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        # TODO: it is possible the entity data version does not match the chunk data version
        ents = self._decode_entity_list(
            self.get_layer_obj(data, self.Entities, pop_last=True)
        ) + self._decode_entity_list(
            self.get_layer_obj(data, self.EntityLayer, pop_last=True)
        )
        if amulet.entity_support:
            chunk.entities = ents
        else:
            chunk._native_entities.extend(ents)
            chunk._native_version = (
                "java",
                self.get_layer_obj(data, self.EntitiesDataVersion),
            )

    def _encode_entities(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        try:
            platform, version = chunk._native_version
        except:
            pass
        else:
            if platform == "java" and isinstance(version, int):
                super()._encode_entities(chunk, data, floor_cy, height_cy)
                self.set_layer_obj(data, self.EntitiesDataVersion, IntTag(version))
                return
        data.pop(self.EntitiesDataVersion[0], None)


export = Anvil2681Interface



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    Tuple,
    Any,
    Generator,
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
    Iterable,
    Callable,
    Type,
    Union,
    TypeVar,
    Generic,
)
import copy
import numpy
import os
import warnings
import logging

import PyMCTranslate

from amulet.api import level as api_level, wrapper as api_wrapper
from amulet.api.chunk import Chunk
from amulet.api.registry import BlockManager
from amulet.api.errors import (
    ChunkLoadError,
    ChunkDoesNotExist,
    ObjectReadError,
    ObjectReadWriteError,
    PlayerDoesNotExist,
    PlayerLoadError,
    EntryLoadError,
    EntryDoesNotExist,
    DimensionDoesNotExist,
)
from amulet.api.data_types import (
    AnyNDArray,
    VersionNumberAny,
    ChunkCoordinates,
    Dimension,
    PlatformType,
)
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.player import Player

if TYPE_CHECKING:
    from amulet.api.wrapper.chunk.translator import Translator

log = logging.getLogger(__name__)

DefaultSelection = SelectionGroup(
    SelectionBox((-30_000_000, 0, -30_000_000), (30_000_000, 256, 30_000_000))
)

VersionNumberT = TypeVar("VersionNumberT", int, Tuple[int, ...])


class FormatWrapper(Generic[VersionNumberT], ABC):
    """
    The FormatWrapper class is a class that sits between the serialised world or structure data and the program using amulet-core.

    It is used to access data from the serialised source in the universal format and write them back again.
    """

    _platform: Optional[PlatformType]
    _version: Optional[VersionNumberT]

    def __init__(self, path: str):
        """
        Construct a new instance of :class:`FormatWrapper`.

        This should not be used directly. You should instead use :func:`amulet.load_format`.

        :param path: The file path to the serialised data.
        """
        self._path = path
        self._is_open = False
        self._has_lock = False
        self._translation_manager = None
        self._platform = None
        self._version = None
        self._bounds: Dict[Dimension, SelectionGroup] = {}
        self._changed: bool = False

    @property
    def sub_chunk_size(self) -> int:
        """
        The dimensions of a sub-chunk.
        """
        return 16

    @property
    def path(self) -> str:
        """The path to the data on disk."""
        return self._path

    @property
    @abstractmethod
    def level_name(self) -> str:
        """The name of the level."""
        raise NotImplementedError

    @property
    def translation_manager(self) -> PyMCTranslate.TranslationManager:
        """The translation manager attached to the world."""
        if self._translation_manager is None:
            self._translation_manager = PyMCTranslate.new_translation_manager()
        return self._translation_manager

    @translation_manager.setter
    def translation_manager(self, value: PyMCTranslate.TranslationManager):
        # TODO: this should not be settable.
        self._translation_manager = value

    @property
    def exists(self) -> bool:
        """Does some data exist at the specified path."""
        return os.path.exists(self.path)

    @staticmethod
    @abstractmethod
    def is_valid(path: str) -> bool:
        """
        Returns whether this format wrapper is able to load the given data.

        :param path: The path of the data to load.
        :return: True if the world can be loaded by this format wrapper, False otherwise.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        """
        The valid platform and version combinations that this object can accept.

        This is used when setting the platform and version in the create_and_open method
        to verify that the platform and version are valid.

        :return: A dictionary mapping the platform to a tuple of two booleans to determine if numerical and blockstate are valid respectively.
        """
        raise NotImplementedError

    @property
    def platform(self) -> PlatformType:
        """Platform string the data is stored in (eg "bedrock" / "java" / ...)"""
        if self._platform is None:
            raise Exception(
                "Cannot access the game platform until the level has been loaded."
            )
        return self._platform

    @property
    def version(self) -> VersionNumberT:
        """The version number for the given platform the data is stored in eg (1, 16, 2)"""
        if self._version is None:
            raise Exception(
                "Cannot access the game version until the level has been loaded."
            )
        return self._version

    @property
    def max_world_version(self) -> Tuple[PlatformType, VersionNumberT]:
        """
        The version the world was last opened in.

        This should be greater than or equal to the chunk versions found within.
        """
        return self.platform, self.version

    @property
    def changed(self) -> bool:
        """Has any data been pushed to the format wrapper that has not been saved to disk."""
        return self._changed

    @property
    @abstractmethod
    def dimensions(self) -> List[Dimension]:
        """A list of all the dimensions contained in the world."""
        raise NotImplementedError

    @property
    @abstractmethod
    def can_add_dimension(self) -> bool:
        """
        Can external code register a new dimension.

        If False :meth:`register_dimension` will have no effect.
        """
        raise NotImplementedError

    @abstractmethod
    def register_dimension(self, dimension_identifier: Any):
        """
        Register a new dimension.

        :param dimension_identifier: The identifier for the dimension.
        """
        raise NotImplementedError

    @property
    def requires_selection(self) -> bool:
        """Does this object require that a selection be defined when creating it from scratch?"""
        return False

    @property
    def multi_selection(self) -> bool:
        """
        Does this object support having multiple selection boxes.

        If False it will be given exactly 1 selection.

        If True can be given 0 or more.
        """
        return False

    @property
    def selection(self) -> SelectionGroup:
        """The area that all chunk data must fit within."""
        warnings.warn(
            "FormatWrapper.selection is depreciated and will be removed in the future. Please use FormatWrapper.bounds(dimension) instead",
            DeprecationWarning,
        )
        return self.bounds(self.dimensions[0])

    def bounds(self, dimension: Dimension) -> SelectionGroup:
        if dimension not in self._bounds:
            if dimension in self.dimensions:
                raise Exception(
                    f'The dimension exists but there is no selection registered for it. Please report this to a developer "{dimension}" {self}'
                )
            else:
                raise DimensionDoesNotExist
        return self._bounds[dimension]

    @abstractmethod
    def _get_interface(
        self, raw_chunk_data: Optional[Any] = None
    ) -> api_wrapper.Interface:
        raise NotImplementedError

    def _get_interface_and_translator(
        self, raw_chunk_data=None
    ) -> Tuple[api_wrapper.Interface, "Translator", VersionNumberAny]:
        interface = self._get_interface(raw_chunk_data)
        translator, version_identifier = interface.get_translator(
            self.max_world_version, raw_chunk_data
        )
        return interface, translator, version_identifier

    def create_and_open(
        self,
        platform: PlatformType,
        version: VersionNumberAny,
        bounds: Union[
            SelectionGroup, Dict[Dimension, Optional[SelectionGroup]], None
        ] = None,
        overwrite: bool = False,
        **kwargs,
    ):
        """
        Remove the data at the path and set up a new database.

        You might want to call :attr:`exists` to check if something exists at the path
        and warn the user they are going to overwrite existing data before calling this method.

        :param platform: The platform the data should use.
        :param version: The version the data should use.
        :param bounds: The bounds for each dimension. If one :class:`SelectionGroup` is given it will be applied to all dimensions.
        :param overwrite: Should an existing database be overwritten. If this is False and one exists and error will be thrown.
        :param kwargs: Extra arguments as each implementation requires.
        :return:
        """
        if self.is_open:
            raise ObjectReadError(f"Cannot open {self} because it was already opened.")

        if (
            platform not in self.valid_formats or len(self.valid_formats[platform]) < 2
        ):  # check that the platform and version are valid
            raise ObjectReadError(
                f"{platform} is not a valid platform for this wrapper."
            )
        translator_version = self.translation_manager.get_version(platform, version)
        if translator_version.has_abstract_format:  # numerical
            if not self.valid_formats[platform][0]:
                raise ObjectReadError(
                    f"The version given ({version}) is from the numerical format but this wrapper does not support the numerical format."
                )
        else:
            if not self.valid_formats[platform][1]:
                raise ObjectReadError(
                    f"The version given ({version}) is from the blockstate format but this wrapper does not support the blockstate format."
                )

        self._platform = translator_version.platform
        self._version = translator_version.version_number
        self._create(overwrite, bounds, **kwargs)

    def _clean_selection(self, selection: SelectionGroup) -> SelectionGroup:
        if self.multi_selection:
            return selection
        else:
            if selection:
                return SelectionGroup(
                    sorted(
                        selection.selection_boxes,
                        reverse=True,
                        key=lambda b: b.volume,
                    )[0]
                )
            else:
                raise ObjectReadError(
                    "A single selection was required but none were given."
                )

    @abstractmethod
    def _create(
        self,
        overwrite: bool,
        bounds: Union[
            SelectionGroup, Dict[Dimension, Optional[SelectionGroup]], None
        ] = None,
        **kwargs,
    ):
        """Set up the database from scratch."""
        raise NotImplementedError

    def open(self):
        """Open the database for reading and writing."""
        if self.is_open:
            raise ObjectReadError(f"Cannot open {self} because it was already opened.")
        self._open()

    @abstractmethod
    def _open(self):
        raise NotImplementedError

    @property
    def is_open(self) -> bool:
        """Has the object been opened."""
        return self._is_open

    @property
    def has_lock(self) -> bool:
        """Verify that the world database can be read and written"""
        return self._has_lock

    def _verify_has_lock(self):
        """
        Ensure that the FormatWrapper is open and has a lock on the object.

        :raises:
            ObjectReadWriteError: if the FormatWrapper does not have a lock on the object.
        """
        if not self.is_open:
            raise ObjectReadWriteError(
                f"The object {self} was never opened. Call .open or .create_and_open to open it before accessing data."
            )
        elif not self.has_lock:
            raise ObjectReadWriteError(
                f"The lock on the object {self} has been lost. It was probably opened somewhere else."
            )

    @staticmethod
    def pre_save_operation(level: api_level.BaseLevel) -> Generator[float, None, bool]:
        """
        Logic to run before saving. Eg recalculating height maps or lighting.

        Must be a generator that yields a number and returns a bool.

        The yielded number is the progress from 0 to 1.

        The returned bool is if changes have been made.

        :param level: The level to apply modifications to.
        :return: Have any modifications been made.
        """
        yield 1
        return False

    def save(self):
        """Save the data back to the level."""
        self._verify_has_lock()
        self._save()
        self._changed = False

    @abstractmethod
    def _save(self):
        raise NotImplementedError

    def close(self):
        """Close the level."""
        if self.is_open:
            self._is_open = False
            self._has_lock = False
            self._close()

    @abstractmethod
    def _close(self):
        raise NotImplementedError

    @abstractmethod
    def unload(self):
        """Unload data stored in the FormatWrapper class"""
        raise NotImplementedError

    @abstractmethod
    def all_chunk_coords(self, dimension: Dimension) -> Iterable[ChunkCoordinates]:
        """A generator of all chunk coords in the given dimension."""
        raise NotImplementedError

    @abstractmethod
    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        """
        Does the chunk exist in the world database?

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the chunk from.
        :return: True if the chunk exists. Calling load_chunk on this chunk may still throw ChunkLoadError
        """
        raise NotImplementedError

    def _safe_load(
        self,
        meth: Callable,
        args: Tuple[Any, ...],
        msg: str,
        load_error: Type[EntryLoadError],
        does_not_exist_error: Type[EntryDoesNotExist],
    ):
        try:
            self._verify_has_lock()
        except ObjectReadWriteError as e:
            raise does_not_exist_error(e)
        try:
            return meth(*args)
        except does_not_exist_error as e:
            raise e
        except Exception as e:
            log.error(msg.format(*args), exc_info=True)
            raise load_error(e) from e

    def load_chunk(self, cx: int, cz: int, dimension: Dimension) -> Chunk:
        """
        Loads and creates a universal :class:`~amulet.api.chunk.Chunk` object from chunk coordinates.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the chunk from.
        :return: The chunk at the given coordinates.
        :raises:
            ChunkDoesNotExist: If the chunk does not exist (was deleted or never created)
            ChunkLoadError: If the chunk was not able to be loaded. Eg. If the chunk is corrupt or some error occurred when loading.
        """
        return self._safe_load(
            self._load_chunk,
            (cx, cz, dimension),
            "Error loading chunk {} {} {}",
            ChunkLoadError,
            ChunkDoesNotExist,
        )

    def _load_chunk(
        self, cx: int, cz: int, dimension: Dimension, recurse: bool = True
    ) -> Chunk:
        """
        Loads and creates a universal :class:`~amulet.api.chunk.Chunk` object from chunk coordinates.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the chunk from.
        :param recurse: bool: look in boundary chunks if required to fully define data
        :return: The chunk at the given coordinates.
        """

        # Gets an interface (the code that actually reads the chunk data)
        raw_chunk_data = self._get_raw_chunk_data(cx, cz, dimension)
        interface, translator, game_version = self._get_interface_and_translator(
            raw_chunk_data
        )

        # decode the raw chunk data into the universal format
        chunk, block_palette = self._decode(
            interface, dimension, cx, cz, raw_chunk_data
        )
        block_palette: AnyNDArray
        chunk = self._unpack(translator, game_version, chunk, block_palette)
        return self._convert_to_load(
            chunk, translator, game_version, dimension, recurse=recurse
        )

    @staticmethod
    def _decode(
        interface: api_wrapper.Interface,
        dimension: Dimension,
        cx: int,
        cz: int,
        raw_chunk_data: Any,
    ) -> Tuple[Chunk, AnyNDArray]:
        return interface.decode(cx, cz, raw_chunk_data)

    def _unpack(
        self,
        translator: "Translator",
        game_version: VersionNumberAny,
        chunk: Chunk,
        block_palette: AnyNDArray,
    ) -> Chunk:
        return translator.unpack(
            game_version, self.translation_manager, chunk, block_palette
        )

    def _convert_to_load(
        self,
        chunk: Chunk,
        translator: "Translator",
        game_version: VersionNumberAny,
        dimension: Dimension,
        recurse: bool = True,
    ) -> Chunk:
        # set up a callback that translator can use to get chunk data
        cx, cz = chunk.cx, chunk.cz
        if recurse:
            chunk_cache: Dict[ChunkCoordinates, Chunk] = {}

            def get_chunk_callback(x: int, z: int) -> Chunk:
                cx_, cz_ = cx + x, cz + z
                if (cx_, cz_) not in chunk_cache:
                    chunk_cache[(cx_, cz_)] = self._load_chunk(
                        cx_, cz_, dimension, recurse=False
                    )
                return chunk_cache[(cx_, cz_)]

        else:
            get_chunk_callback = None

        # translate the data to universal format
        chunk = translator.to_universal(
            game_version, self.translation_manager, chunk, get_chunk_callback, recurse
        )

        chunk.changed = False
        return chunk

    def commit_chunk(self, chunk: Chunk, dimension: Dimension):
        """
        Save a universal format chunk to the FormatWrapper database (not the level database)

        Call save method to write changed chunks back to the level.

        :param chunk: The chunk object to translate and save.
        :param dimension: The dimension to commit the chunk to.
        """
        try:
            self._verify_has_lock()
        except ObjectReadWriteError as e:
            log.error(e)
        try:
            self._commit_chunk(copy.deepcopy(chunk), dimension)
        except Exception:
            log.error(f"Error saving chunk {chunk}", exc_info=True)
        self._changed = True

    def _commit_chunk(self, chunk: Chunk, dimension: Dimension, recurse: bool = True):
        """
        Saves a universal :class:`~amulet.api.chunk.Chunk` object.

        Calls the interface then the translator.

        It then calls _put_chunk_data to store the data returned by the interface
        """
        # get the coordinates for later
        cx, cz = chunk.cx, chunk.cz

        # Gets an interface, translator and most recent chunk version for the game version.
        interface, translator, chunk_version = self._get_interface_and_translator()

        chunk = self._convert_to_save(chunk, chunk_version, translator, recurse)
        chunk, chunk_palette = self._pack(chunk, translator, chunk_version)
        raw_chunk_data = self._encode(interface, chunk, dimension, chunk_palette)

        self._put_raw_chunk_data(cx, cz, raw_chunk_data, dimension)

    def _convert_to_save(
        self,
        chunk: Chunk,
        chunk_version: VersionNumberAny,
        translator: "Translator",
        recurse: bool = True,
    ) -> Chunk:
        """Convert the Chunk in Universal format to a Chunk in the version specific format."""
        # create a new streamlined block block_palette and remap the data
        palette: List[numpy.ndarray] = []
        palette_len = 0
        for cy in chunk.blocks.sub_chunks:
            sub_chunk_palette, sub_chunk = numpy.unique(
                chunk.blocks.get_sub_chunk(cy), return_inverse=True
            )
            chunk.blocks.add_sub_chunk(
                cy, sub_chunk.astype(numpy.uint32).reshape((16, 16, 16)) + palette_len
            )
            palette_len += len(sub_chunk_palette)
            palette.append(sub_chunk_palette)

        if palette:
            chunk_palette, lut = numpy.unique(
                numpy.concatenate(palette), return_inverse=True
            )
            for cy in chunk.blocks.sub_chunks:
                chunk.blocks.add_sub_chunk(
                    cy, lut.astype(numpy.uint32)[chunk.blocks.get_sub_chunk(cy)]
                )
            chunk._block_palette = BlockManager(
                numpy.vectorize(chunk.block_palette.__getitem__)(chunk_palette)
            )
        else:
            chunk._block_palette = BlockManager()

        def get_chunk_callback(_: int, __: int) -> Chunk:
            # conversion from universal should not require any data outside the block
            return chunk

        # translate from universal format to version format
        return translator.from_universal(
            chunk_version, self.translation_manager, chunk, get_chunk_callback, recurse
        )

    def _pack(
        self, chunk: Chunk, translator: "Translator", chunk_version: VersionNumberAny
    ) -> Tuple[Chunk, AnyNDArray]:
        """Pack the chunk data into the format required by the encoder.
        This includes converting the string names to numerical formats for the versions that require it.
        """
        return translator.pack(chunk_version, self.translation_manager, chunk)

    def _encode(
        self,
        interface: api_wrapper.Interface,
        chunk: Chunk,
        dimension: Dimension,
        chunk_palette: AnyNDArray,
    ) -> Any:
        """Encode the data to the raw format as saved on disk."""
        raise NotImplementedError

    def delete_chunk(self, cx: int, cz: int, dimension: Dimension):
        """
        Delete the given chunk from the level.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        """
        self._delete_chunk(cx, cz, dimension)
        self._changed = True

    @abstractmethod
    def _delete_chunk(self, cx: int, cz: int, dimension: Dimension):
        raise NotImplementedError

    def put_raw_chunk_data(self, cx: int, cz: int, data: Any, dimension: Dimension):
        """
        Commit the raw chunk data to the FormatWrapper cache.

        Call :meth:`save` to push all the cache data to the level.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param data: The raw data to commit to the level.
        :param dimension: The dimension to load the data from.
        """
        self._verify_has_lock()
        self._put_raw_chunk_data(cx, cz, data, dimension)

    @abstractmethod
    def _put_raw_chunk_data(self, cx: int, cz: int, data: Any, dimension: Dimension):
        raise NotImplementedError

    def get_raw_chunk_data(self, cx: int, cz: int, dimension: Dimension) -> Any:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        return self._safe_load(
            self._get_raw_chunk_data,
            (cx, cz, dimension),
            "Error loading chunk {} {} {}",
            ChunkLoadError,
            ChunkDoesNotExist,
        )

    @abstractmethod
    def _get_raw_chunk_data(self, cx: int, cz: int, dimension: Dimension) -> Any:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        raise NotImplementedError

    @abstractmethod
    def all_player_ids(self) -> Iterable[str]:
        """
        Returns a set of all player ids that are present in the level
        """
        return NotImplemented

    @abstractmethod
    def has_player(self, player_id: str) -> bool:
        """
        Test if a player id is present in the level.
        """
        return NotImplemented

    def load_player(self, player_id: str) -> "Player":
        """
        Gets the :class:`Player` object that belongs to the specified player id

        If no parameter is supplied, the data of the local player should be returned

        :param player_id: The desired player id
        :return: A Player instance
        """
        return self._safe_load(
            self._load_player,
            (player_id,),
            "Error loading player {}",
            PlayerLoadError,
            PlayerDoesNotExist,
        )

    @abstractmethod
    def _load_player(self, player_id: str) -> "Player":
        """
        Get the raw player data and unpack it into a Player class.

        :param player_id: The id of the player to get.
        :return:
        """
        raise NotImplementedError

    def get_raw_player_data(self, player_id: str) -> Any:
        """
        Get the player data in the lowest level form.

        :param player_id: The id of the player to get.
        :return:
        """
        return self._safe_load(
            self._get_raw_player_data,
            (player_id,),
            "Error loading player {}",
            PlayerLoadError,
            PlayerDoesNotExist,
        )

    @abstractmethod
    def _get_raw_player_data(self, player_id: str) -> Any:
        raise NotImplementedError



from __future__ import annotations

import os
import struct
import warnings
from typing import Tuple, Dict, Union, Optional, List, BinaryIO, Iterable, Any
from io import BytesIO
import shutil
import traceback
import time

from amulet_nbt import (
    AbstractBaseTag,
    load as load_nbt,
    NamedTag,
    CompoundTag,
    StringTag,
    ByteTag,
    ShortTag,
    IntTag,
    ListTag,
    LongTag,
    FloatTag,
    utf8_escape_decoder,
    utf8_escape_encoder,
    ReadContext,
)
from amulet.api.player import Player, LOCAL_PLAYER
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionBox, SelectionGroup

from leveldb import LevelDB, LevelDBException, LevelDBEncrypted
from amulet.utils.format_utils import check_all_exist
from amulet.api.data_types import (
    ChunkCoordinates,
    VersionNumberTuple,
    PlatformType,
    Dimension,
    AnyNDArray,
)
from amulet.api.wrapper import WorldFormatWrapper
from amulet.api.errors import ObjectWriteError, ObjectReadError, PlayerDoesNotExist

from .interface.chunk.leveldb_chunk_versions import (
    game_to_chunk_version,
)
from .dimension import (
    LevelDBDimensionManager,
    ChunkData,
    OVERWORLD,
    THE_NETHER,
    THE_END,
)
from .interface.chunk import BaseLevelDBInterface, get_interface

InternalDimension = Optional[int]


class BedrockLevelDAT(NamedTag):
    _path: str
    _level_dat_version: int

    def __init__(
        self, tag=None, name: str = "", path: str = None, level_dat_version: int = None
    ):
        if isinstance(tag, str):
            warnings.warn(
                "You must use BedrockLevelDAT.from_file to load from a file.",
                FutureWarning,
            )
            super().__init__()
            self._path = path = tag
            self._level_dat_version = 8
            if os.path.isfile(path):
                self.load_from(path)
            return
        else:
            if not (isinstance(path, str) and isinstance(level_dat_version, int)):
                raise TypeError(
                    "path and level_dat_version must be specified when constructing a BedrockLevelDAT instance."
                )
            super().__init__(tag, name)
            self._path = path
            self._level_dat_version = level_dat_version

    @classmethod
    def from_file(cls, path: str):
        level_dat_version, name, tag = cls._read_from(path)
        return cls(tag, name, path, level_dat_version)

    @property
    def path(self) -> Optional[str]:
        return self._path

    @staticmethod
    def _read_from(path: str) -> Tuple[int, str, AbstractBaseTag]:
        with open(path, "rb") as f:
            level_dat_version = struct.unpack("<i", f.read(4))[0]
            if 4 <= level_dat_version <= 10:
                data_length = struct.unpack("<i", f.read(4))[0]
                root_tag = load_nbt(
                    f.read(data_length),
                    compressed=False,
                    little_endian=True,
                    string_decoder=utf8_escape_decoder,
                )
                name = root_tag.name
                value = root_tag.tag
            else:
                # TODO: handle other versions
                raise ObjectReadError(
                    f"Unsupported level.dat version {level_dat_version}"
                )
        return level_dat_version, name, value

    def load_from(self, path: str):
        self._level_dat_version, self.name, self.tag = self._read_from(path)

    def reload(self):
        self.load_from(self.path)

    def save(self, path: str = None):
        self.save_to(path or self._path)

    def save_to(
        self,
        filename_or_buffer: Union[str, BinaryIO] = None,
        *,
        compressed=False,
        little_endian=True,
        string_encoder=utf8_escape_encoder,
    ) -> Optional[bytes]:
        payload = super().save_to(
            compressed=compressed,
            little_endian=little_endian,
            string_encoder=string_encoder,
        )
        buffer = BytesIO()
        buffer.write(struct.pack("<ii", self._level_dat_version, len(payload)))
        buffer.write(payload)
        if filename_or_buffer is None:
            return buffer.getvalue()
        elif isinstance(filename_or_buffer, str):
            with open(filename_or_buffer, "wb") as f:
                f.write(buffer.getvalue())
        else:
            filename_or_buffer.write(buffer.getvalue())


class LevelDBFormat(WorldFormatWrapper[VersionNumberTuple]):
    """
    This FormatWrapper class exists to interface with the Bedrock world format.
    """

    # The leveldb database. Access it through the public property `level_db`
    _db: Optional[LevelDB]
    # A class to manage dimension data. This is private
    _dimension_manager: Optional[LevelDBDimensionManager]

    _root_tag: BedrockLevelDAT

    def __init__(self, path: str):
        """
        Construct a new instance of :class:`LevelDBFormat`.

        This should not be used directly. You should instead use :func:`amulet.load_format`.

        :param path: The file path to the serialised data.
        """
        super().__init__(path)
        self._platform = "bedrock"
        dat_path = os.path.join(path, "level.dat")
        if os.path.isfile(dat_path):
            self._root_tag = BedrockLevelDAT.from_file(dat_path)
        else:
            # TODO: handle level creation better
            self._root_tag = BedrockLevelDAT(path=dat_path, level_dat_version=9)
        self._db = None
        self._dimension_manager = None
        self._shallow_load()

    def _shallow_load(self):
        try:
            self._load_level_dat()
        except:
            pass

    def _load_level_dat(self):
        """Load the level.dat file and check the image file"""
        if os.path.isfile(os.path.join(self.path, "world_icon.jpeg")):
            self._world_image_path = os.path.join(self.path, "world_icon.jpeg")
        self.root_tag = BedrockLevelDAT.from_file(os.path.join(self.path, "level.dat"))

    @staticmethod
    def is_valid(path: str):
        return check_all_exist(path, "db", "level.dat", "levelname.txt")

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {"bedrock": (True, True)}

    @property
    def version(self) -> VersionNumberTuple:
        if self._version is None:
            self._version = self._get_version()
        return self._version

    def _get_version(self) -> VersionNumberTuple:
        """
        The version the world was last opened in.

        This should be greater than or equal to the chunk versions found within

        For this format wrapper it returns a tuple of 3/4 ints (the game version number)
        """
        try:
            return tuple(
                [
                    t.py_int
                    for t in self.root_tag.compound.get_list("lastOpenedWithVersion")
                ]
            )
        except Exception:
            return 1, 2, 0

    @property
    def root_tag(self) -> BedrockLevelDAT:
        """The level.dat data for the level."""
        return self._root_tag

    @root_tag.setter
    def root_tag(self, root_tag: Union[NamedTag, CompoundTag, BedrockLevelDAT]):
        if isinstance(root_tag, CompoundTag):
            self._root_tag.tag = root_tag
        elif isinstance(root_tag, NamedTag):
            self._root_tag.name = root_tag.name
            self._root_tag.tag = root_tag.compound
        else:
            raise ValueError(
                "root_tag must be a CompoundTag, NamedTag or BedrockLevelDAT"
            )

    @property
    def level_name(self) -> str:
        return self.root_tag.compound.get_string("LevelName", StringTag()).py_str

    @level_name.setter
    def level_name(self, value: str):
        self.root_tag.compound["LevelName"] = StringTag(value)

    @property
    def last_played(self) -> int:
        return self.root_tag.compound.get_long("LastPlayed", LongTag()).py_int

    @property
    def game_version_string(self) -> str:
        try:
            return f'Bedrock {".".join(str(v.py_int) for v in self.root_tag.compound.get_list("lastOpenedWithVersion"))}'
        except Exception:
            return f"Bedrock Unknown Version"

    @property
    def dimensions(self) -> List["Dimension"]:
        self._verify_has_lock()
        return self._dimension_manager.dimensions

    # def register_dimension(
    #     self, dimension_internal: int, dimension_name: Optional["Dimension"] = None
    # ):
    #     """
    #     Register a new dimension.
    #
    #     :param dimension_internal: The internal integer representation of the dimension.
    #     :param dimension_name: The name of the dimension shown to the user.
    #     :return:
    #     """
    #     self._dimension_manager.register_dimension(dimension_internal, dimension_name)

    @property
    def level_db(self) -> LevelDB:
        """The raw leveldb database."""
        if self._db is None:
            raise Exception(
                "The world is not open. The leveldb database cannot be accessed."
            )
        return self._db

    @property
    def _level_manager(self) -> LevelDBDimensionManager:
        warnings.warn(
            "_level_manager attribute is depreciated. If you want to access the raw leveldb database it can be accessed through the level_db property."
        )
        return self._dimension_manager

    def _get_interface(
        self, raw_chunk_data: Optional[Any] = None
    ) -> BaseLevelDBInterface:
        return get_interface(self._get_interface_key(raw_chunk_data))

    def _get_interface_key(self, raw_chunk_data: Optional[ChunkData] = None) -> int:
        if raw_chunk_data:
            if b"," in raw_chunk_data:
                chunk_version = raw_chunk_data[b","][0]
            else:
                chunk_version = raw_chunk_data.get(b"v", b"\x00")[0]
        else:
            chunk_version = game_to_chunk_version(
                self.max_world_version[1],
                self.root_tag.compound.get_compound("experiments", CompoundTag())
                .get_byte("caves_and_cliffs", ByteTag())
                .py_int,
            )
        return chunk_version

    def _decode(
        self,
        interface: BaseLevelDBInterface,
        dimension: Dimension,
        cx: int,
        cz: int,
        raw_chunk_data: Any,
    ) -> Tuple[Chunk, AnyNDArray]:
        bounds = self.bounds(dimension).bounds
        return interface.decode(cx, cz, raw_chunk_data, (bounds[0][1], bounds[1][1]))

    def _encode(
        self,
        interface: BaseLevelDBInterface,
        chunk: Chunk,
        dimension: Dimension,
        chunk_palette: AnyNDArray,
    ) -> Any:
        bounds = self.bounds(dimension).bounds
        return interface.encode(
            chunk,
            chunk_palette,
            self.max_world_version,
            (bounds[0][1], bounds[1][1]),
        )

    def _reload_world(self):
        try:
            self.close()
        except:
            pass
        try:
            self._db = LevelDB(os.path.join(self.path, "db"))
            self._dimension_manager = LevelDBDimensionManager(self)
            self._is_open = True
            self._has_lock = True
            experiments = self.root_tag.compound.get_compound(
                "experiments", CompoundTag()
            )
            if (
                experiments.get_byte("caves_and_cliffs", ByteTag()).py_int
                or experiments.get_byte("caves_and_cliffs_internal", ByteTag()).py_int
                or self.version >= (1, 18)
            ):
                self._bounds[OVERWORLD] = SelectionGroup(
                    SelectionBox(
                        (-30_000_000, -64, -30_000_000), (30_000_000, 320, 30_000_000)
                    )
                )
            else:
                self._bounds[OVERWORLD] = SelectionGroup(
                    SelectionBox(
                        (-30_000_000, 0, -30_000_000), (30_000_000, 256, 30_000_000)
                    )
                )
            self._bounds[THE_NETHER] = SelectionGroup(
                SelectionBox(
                    (-30_000_000, 0, -30_000_000), (30_000_000, 128, 30_000_000)
                )
            )
            self._bounds[THE_END] = SelectionGroup(
                SelectionBox(
                    (-30_000_000, 0, -30_000_000), (30_000_000, 256, 30_000_000)
                )
            )
            if b"LevelChunkMetaDataDictionary" in self.level_db:
                data = self.level_db[b"LevelChunkMetaDataDictionary"]
                count, data = struct.unpack("<I", data[:4])[0], data[4:]
                for _ in range(count):
                    key, data = data[:8], data[8:]
                    context = ReadContext()
                    value = load_nbt(
                        data,
                        little_endian=True,
                        compressed=False,
                        string_decoder=utf8_escape_decoder,
                        read_context=context,
                    ).compound
                    data = data[context.offset :]

                    try:
                        dimension_name = value.get_string("DimensionName").py_str
                        # The dimension names are stored differently TODO: split local and global names
                        dimension_name = {
                            "Overworld": "minecraft:overworld",
                            "Nether": "minecraft:the_nether",
                            "TheEnd": "minecraft:the_end",
                        }.get(dimension_name, dimension_name)

                    except KeyError:
                        # Some entries seem to not have a dimension assigned to them. Is there a default? We will skip over these for now.
                        # {'LastSavedBaseGameVersion': StringTag("1.19.81"), 'LastSavedDimensionHeightRange': CompoundTag({'max': ShortTag(320), 'min': ShortTag(-64)})}
                        pass
                    else:
                        previous_bounds = self._bounds.get(
                            dimension_name,
                            SelectionGroup(
                                SelectionBox(
                                    (-30_000_000, 0, -30_000_000),
                                    (30_000_000, 256, 30_000_000),
                                )
                            ),
                        )
                        min_y = min(
                            value.get_compound(
                                "LastSavedDimensionHeightRange", CompoundTag()
                            )
                            .get_short("min", ShortTag())
                            .py_int,
                            value.get_compound(
                                "OriginalDimensionHeightRange", CompoundTag()
                            )
                            .get_short("min", ShortTag())
                            .py_int,
                            previous_bounds.min_y,
                        )
                        max_y = max(
                            value.get_compound(
                                "LastSavedDimensionHeightRange", CompoundTag()
                            )
                            .get_short("max", ShortTag())
                            .py_int,
                            value.get_compound(
                                "OriginalDimensionHeightRange", CompoundTag()
                            )
                            .get_short("max", ShortTag())
                            .py_int,
                            previous_bounds.max_y,
                        )
                        self._bounds[dimension_name] = SelectionGroup(
                            SelectionBox(
                                (previous_bounds.min_x, min_y, previous_bounds.min_z),
                                (previous_bounds.max_x, max_y, previous_bounds.max_z),
                            )
                        )
        except LevelDBEncrypted as e:
            self._is_open = self._has_lock = False
            raise LevelDBException(
                "It looks like this world is from the marketplace.\nThese worlds are encrypted and cannot be edited."
            ) from e
        except LevelDBException as e:
            msg = str(e)
            self._is_open = self._has_lock = False
            # I don't know if there is a better way of handling this.
            if msg.startswith("IO error:") and msg.endswith(": Permission denied"):
                traceback.print_exc()
                raise LevelDBException(
                    f"Failed to load the database. The world may be open somewhere else.\n{msg}"
                ) from e
            else:
                raise e

    def _open(self):
        """Open the database for reading and writing"""
        self._reload_world()

    def _create(
        self,
        overwrite: bool,
        bounds: Union[
            SelectionGroup, Dict[Dimension, Optional[SelectionGroup]], None
        ] = None,
        **kwargs,
    ):
        if os.path.isdir(self.path):
            if overwrite:
                shutil.rmtree(self.path)
            else:
                raise ObjectWriteError(
                    f"A world already exists at the path {self.path}"
                )

        version = self.translation_manager.get_version(
            self.platform, self.version
        ).version_number
        self._version = version + (0,) * (5 - len(version))

        self.root_tag = root = CompoundTag()
        root["StorageVersion"] = IntTag(8)
        root["lastOpenedWithVersion"] = ListTag([IntTag(i) for i in self._version])
        root["Generator"] = IntTag(1)
        root["LastPlayed"] = LongTag(int(time.time()))
        root["LevelName"] = StringTag("World Created By Amulet")

        os.makedirs(self.path, exist_ok=True)
        self.root_tag.save()

        db = LevelDB(os.path.join(self.path, "db"), True)
        db.close()

        self._reload_world()

    @property
    def has_lock(self) -> bool:
        if self._has_lock:
            return True  # TODO: implement a check to ensure access to the database
        return False

    def _save(self):
        os.makedirs(self.path, exist_ok=True)
        self.root_tag.save()
        with open(os.path.join(self.path, "levelname.txt"), "w", encoding="utf-8") as f:
            f.write(self.level_name)

    def _close(self):
        self._db.close()
        self._db = None
        self._dimension_manager = None
        self._actor_counter = None

    def unload(self):
        pass

    def all_chunk_coords(self, dimension: "Dimension") -> Iterable[ChunkCoordinates]:
        self._verify_has_lock()
        yield from self._dimension_manager.all_chunk_coords(dimension)

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return self._dimension_manager.has_chunk(cx, cz, dimension)

    def _delete_chunk(self, cx: int, cz: int, dimension: "Dimension"):
        self._dimension_manager.delete_chunk(cx, cz, dimension)

    def _put_raw_chunk_data(
        self, cx: int, cz: int, data: ChunkData, dimension: "Dimension"
    ):
        return self._dimension_manager.put_chunk_data(cx, cz, data, dimension)

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: "Dimension"
    ) -> ChunkData:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        return self._dimension_manager.get_chunk_data(cx, cz, dimension)

    def all_player_ids(self) -> Iterable[str]:
        """
        Returns a generator of all player ids that are present in the level
        """
        yield from (
            pid[7:].decode("utf-8")
            for pid, _ in self._db.iterate(b"player_", b"player_\xFF")
        )
        if self.has_player(LOCAL_PLAYER):
            yield LOCAL_PLAYER

    def has_player(self, player_id: str) -> bool:
        if player_id != LOCAL_PLAYER:
            player_id = f"player_{player_id}"
        return player_id.encode("utf-8") in self._db

    def _load_player(self, player_id: str) -> Player:
        """
        Gets the :class:`Player` object that belongs to the specified player id

        If no parameter is supplied, the data of the local player will be returned

        :param player_id: The desired player id
        :return: A Player instance
        """
        player_nbt = self._get_raw_player_data(player_id).compound
        dimension = player_nbt["DimensionId"]
        if isinstance(dimension, IntTag) and IntTag(0) <= dimension <= IntTag(2):
            dimension_str = {
                0: OVERWORLD,
                1: THE_NETHER,
                2: THE_END,
            }[dimension.py_int]
        else:
            dimension_str = OVERWORLD

        # get the players position
        pos_data = player_nbt.get("Pos")
        if (
            isinstance(pos_data, ListTag)
            and len(pos_data) == 3
            and pos_data.list_data_type == FloatTag.tag_id
        ):
            position = tuple(map(float, pos_data))
            position = tuple(
                p if -100_000_000 <= p <= 100_000_000 else 0.0 for p in position
            )
        else:
            position = (0.0, 0.0, 0.0)

        # get the players rotation
        rot_data = player_nbt.get("Rotation")
        if (
            isinstance(rot_data, ListTag)
            and len(rot_data) == 2
            and rot_data.list_data_type == FloatTag.tag_id
        ):
            rotation = tuple(map(float, rot_data))
            rotation = tuple(
                p if -100_000_000 <= p <= 100_000_000 else 0.0 for p in rotation
            )
        else:
            rotation = (0.0, 0.0)

        return Player(
            player_id,
            dimension_str,
            position,
            rotation,
        )

    def _get_raw_player_data(self, player_id: str) -> NamedTag:
        if player_id == LOCAL_PLAYER:
            key = player_id.encode("utf-8")
        else:
            key = f"player_{player_id}".encode("utf-8")
        try:
            data = self._db.get(key)
        except KeyError:
            raise PlayerDoesNotExist(f"Player {player_id} doesn't exist")
        return load_nbt(
            data,
            compressed=False,
            little_endian=True,
            string_decoder=utf8_escape_decoder,
        )



from typing import Iterable, BinaryIO
from .data_pack import DataPack


class DataPackManager:
    """
    The DataPackManager class contains one or more data packs.
    It manages loading them so that the stacking order is maintained.
    """

    def __init__(self, data_packs: Iterable[DataPack]):
        """
        Construct a new DataPackManager class.

        :param data_packs: The data packs to load from. Later in the list get higher priority.
        """
        self._data_packs = tuple(
            reversed(
                tuple(
                    pack
                    for pack in data_packs
                    if isinstance(pack, DataPack) and pack.is_valid
                )
            )
        )

    @property
    def all_files(self) -> Iterable[str]:
        """
        The relative paths of all files contained within.

        :return: An iterable of paths.
        """
        all_files = set()
        for pack in self._data_packs:
            all_files.update(pack.all_files)
        return all_files

    def all_files_match(self, match: str) -> Iterable[str]:
        """
        The relative paths of all files contained within that match the given regex.

        :param match: The regex string to match again.
        :return: An iterable of file paths that match the given regex.
        """
        all_files = set()
        for pack in self._data_packs:
            all_files.update(pack.all_files_match(match))
        return all_files

    def has_file(self, relative_path: str) -> bool:
        """
        Does the requested file exist.

        :param relative_path:
        :return:
        """
        return any(pack.has_file(relative_path) for pack in self._data_packs)

    def open(self, relative_path: str) -> BinaryIO:
        """
        Get the contents of the file.

        :param relative_path:
        :return:
        """
        for pack in self._data_packs:
            if pack.has_file(relative_path):
                return pack.open(relative_path)
        raise FileNotFoundError(
            f"The requested path {relative_path} could not be found."
        )

    def close(self):
        """Close the contents."""
        for pack in self._data_packs:
            pack.close()



from .world_utils import (
    SECTOR_BYTES,
    SECTOR_INTS,
    CHUNK_HEADER_SIZE,
    VERSION_GZIP,
    VERSION_DEFLATE,
    block_coords_to_chunk_coords,
    chunk_coords_to_region_coords,
    region_coords_to_chunk_coords,
    blocks_slice_to_chunk_slice,
    gunzip,
    from_nibble_array,
)
from .format_utils import *



from typing import List, Union, Type, Dict
import numpy

from amulet_nbt import (
    IntTag,
    DoubleTag,
    StringTag,
    ListTag,
    CompoundTag,
    ByteArrayTag,
    IntArrayTag,
    LongArrayTag,
    NamedTag,
)

from amulet.api.block import Block
from amulet.api.entity import Entity
from amulet.api.block_entity import BlockEntity
from amulet.api.registry import BlockManager


def unpack_palette(raw_palette: ListTag) -> List[Block]:
    block_palette = []
    extra_block_map: Dict[int, ListTag[IntTag]] = {}
    for block_index, block_nbt in enumerate(raw_palette):
        block_nbt: CompoundTag
        block = Block(
            namespace=block_nbt.get_string("namespace").py_str,
            base_name=block_nbt.get_string("blockname").py_str,
            properties=block_nbt.get_compound("properties").py_dict,
        )

        if block_nbt.get_list("extra_blocks"):
            extra_block_map[block_index] = block_nbt.get_list("extra_blocks")

        block_palette.append(block)

    extra_blocks: ListTag[IntTag]
    for block_index, extra_blocks in extra_block_map.items():
        extra_block_objects = [block_palette[i.py_int] for i in extra_blocks]

        resulting_block = block_palette[block_index]
        for extra_block in extra_block_objects:
            resulting_block = resulting_block + extra_block

        block_palette[block_index] = resulting_block
    return block_palette


def parse_entities(entities: ListTag) -> List[Entity]:
    return [
        Entity(
            entity.get_string("namespace").py_str,
            entity.get_string("base_name").py_str,
            entity.get_double("x").py_float,
            entity.get_double("y").py_float,
            entity.get_double("z").py_float,
            NamedTag(entity.get_compound("nbt")),
        )
        for entity in entities
    ]


def parse_block_entities(block_entities: ListTag) -> List[BlockEntity]:
    return [
        BlockEntity(
            block_entity.get_string("namespace").py_str,
            block_entity.get_string("base_name").py_str,
            block_entity.get_int("x").py_int,
            block_entity.get_int("y").py_int,
            block_entity.get_int("z").py_int,
            NamedTag(block_entity.get_compound("nbt")),
        )
        for block_entity in block_entities
    ]


def generate_block_entry(block: Block, palette_len, extra_blocks) -> CompoundTag:
    return CompoundTag(
        {
            "namespace": StringTag(block.namespace),
            "blockname": StringTag(block.base_name),
            "properties": CompoundTag(block.properties),
            "extra_blocks": ListTag(
                [
                    IntTag(palette_len + extra_blocks.index(_extra_block))
                    for _extra_block in block.extra_blocks
                ]
            ),
        }
    )


def serialise_entities(entities: List[Entity]) -> ListTag:
    return ListTag(
        [
            CompoundTag(
                {
                    "namespace": StringTag(entity.namespace),
                    "base_name": StringTag(entity.base_name),
                    "x": DoubleTag(entity.x),
                    "y": DoubleTag(entity.y),
                    "z": DoubleTag(entity.z),
                    "nbt": entity.nbt.compound,
                }
            )
            for entity in entities
        ]
    )


def serialise_block_entities(
    block_entities: List[BlockEntity],
) -> ListTag:
    return ListTag(
        [
            CompoundTag(
                {
                    "namespace": StringTag(block_entity.namespace),
                    "base_name": StringTag(block_entity.base_name),
                    "x": IntTag(block_entity.x),
                    "y": IntTag(block_entity.y),
                    "z": IntTag(block_entity.z),
                    "nbt": block_entity.nbt.compound,
                }
            )
            for block_entity in block_entities
        ]
    )


def find_fitting_array_type(
    array: numpy.ndarray,
) -> Union[Type[IntArrayTag], Type[ByteArrayTag], Type[LongArrayTag]]:
    max_element = array.max(initial=0)

    if max_element <= 127:
        return ByteArrayTag
    elif max_element <= 2_147_483_647:
        return IntArrayTag
    else:
        return LongArrayTag


def pack_palette(palette: BlockManager) -> ListTag:
    block_palette_nbt = ListTag()
    extra_blocks = set()

    for block in palette.blocks:
        if len(block.extra_blocks) > 0:
            extra_blocks.update(block.extra_blocks)

    extra_blocks = list(extra_blocks)

    palette_len = len(palette)
    for block_entry in palette.blocks:
        block_palette_nbt.append(
            generate_block_entry(block_entry, palette_len, extra_blocks)
        )

    for extra_block in extra_blocks:
        block_palette_nbt.append(
            generate_block_entry(extra_block, palette_len, extra_blocks)
        )
    return block_palette_nbt



from __future__ import annotations

import numpy
from typing import Tuple, Union
from amulet_nbt import NamedTag
from .abstract_base_entity import AbstractBaseEntity


class Entity(AbstractBaseEntity):
    """
    A class to contain all the data to define an Entity.
    """

    obj_name = "Entity"
    coord_types = (float, numpy.floating)

    def __init__(
        self,
        namespace: str,
        base_name: str,
        x: float,
        y: float,
        z: float,
        nbt: NamedTag,
    ):
        """
        Constructs a :class:`Entity` instance.

        :param namespace: The namespace of the entity eg "minecraft"
        :param base_name: The base name of the entity eg "creeper"
        :param x: The x coordinate of the entity
        :param y: The y coordinate of the entity
        :param z: The z coordinate of the entity
        :param nbt: The NBT stored with the entity
        """
        super().__init__(namespace, base_name, x, y, z, nbt)

    @property
    def x(self) -> float:
        """The x location of the Entity."""
        return self._x

    @x.setter
    def x(self, value: float):
        self._x = value

    @property
    def y(self) -> float:
        """The y location of the Entity."""
        return self._y

    @y.setter
    def y(self, value: float):
        self._y = value

    @property
    def z(self) -> float:
        """The z location of the Entity."""
        return self._z

    @z.setter
    def z(self, value: float):
        self._z = value

    @property
    def location(self) -> Tuple[float, float, float]:
        """The location of the Entity."""
        return self._x, self._y, self._z

    @location.setter
    def location(
        self, location: Tuple[Union[int, float], Union[int, float], Union[int, float]]
    ):
        self._x, self._y, self._z = location



from typing import Union, Dict, List, Tuple
from enum import Enum
import logging

log = logging.getLogger(__name__)


class StatusFormats(Enum):
    """
    The different formats the status value can be converted to.

    >>> StatusFormats.Raw  # The raw float value
    >>> StatusFormats.Bedrock  # as an integer between 0 and 2
    >>> StatusFormats.Java_13  # as a string
    >>> StatusFormats.Java_14  # as a string
    """

    Raw = 0  # The raw float value
    Bedrock = 1  # as an integer between 0 and 2
    Java_13 = 2  # as a string
    Java_14 = 3  # as a string
    Java_20 = 20  # as a string


J13 = StatusFormats.Java_13
J14 = StatusFormats.Java_14
J20 = StatusFormats.Java_20

states = {
    "empty": [[J13, J14], -1],
    "minecraft:empty": [[J20], -1],
    "structure_starts": [[J14], -0.9],
    "minecraft:structure_starts": [[J20], -0.9],
    "structure_references": [[J14], -0.8],
    "minecraft:structure_references": [[J20], -0.8],
    "biomes": [[J14], -0.7],
    "minecraft:biomes": [[J20], -0.7],
    "noise": [[J14], -0.6],
    "minecraft:noise": [[J20], -0.6],
    "base": [[J13], -0.5],
    "surface": [[J14], -0.5],
    "minecraft:surface": [[J20], -0.5],
    "carved": [[J13], -0.4],
    "carvers": [[J14], -0.4],
    "minecraft:carvers": [[J20], -0.4],
    "liquid_carved": [[J13], -0.3],
    "liquid_carvers": [[J14], -0.3],
    "decorated": [[J13], -0.2],
    "features": [[J14], -0.2],
    "minecraft:features": [[J20], -0.2],
    "lighted": [[J13], -0.1],
    "minecraft:initialize_light": [[J20], -0.1],
    "light": [[J14], -0.1],
    "minecraft:light": [[J20], -0.1],
    # 0.0	needs ticked
    # 1.0	needs population
    "mobs_spawned": [[J13], 1.1],
    "spawn": [[J14], 1.1],
    "minecraft:spawn": [[J20], 1.1],
    "finalized": [[J13], 1.5],
    "heightmaps": [[J14], 1.5],
    "fullchunk": [[J13], 1.9],
    "postprocessed": [[J13], 2.0],
    "full": [[J14], 2.0],
    "minecraft:full": [[J20], 2.0],
    # 2.0	done
}

versions: Dict[StatusFormats, List[Tuple[float, str]]] = {}
for key, val in states.items():
    for version_ in val[0]:
        versions.setdefault(version_, []).append((val[1], key))
for data in versions.values():
    data.sort(key=lambda val_: val_[0], reverse=True)


class Status:
    """
    A class to represent the generation state of a chunk.

    This stores the chunk status as a float in the range -1 to 2 where -1 is nothing generated and 2 being fully generated.
    """

    def __init__(self):
        """
        Construct an instance of the :class:`Status` class.
        """
        self._value = 2.0

    @property
    def value(self) -> float:
        """
        **Getter**

        Get the raw status value.

        >>> status = chunk.status.value

        :return: The float status value in the range -1 to 2

        **Setter**

        Set the status value.

        >>> chunk.status.value = 2
        >>> chunk.status.value = "full"

        This can be a float/int or a string. If it is a string it is looked up in the states look up table to find the float value it corresponds with.

        :param value: The value to set as the generation stage.
        """
        return self._value

    @value.setter
    def value(self, value: Union[float, int, str]):
        """
        Set the status value.

        >>> chunk.status.value = 2
        >>> chunk.status.value = "full"

        This can be a float/int or a string. If it is a string it is looked up in the states look up table to find the float value it corresponds with.

        :param value: The value to set as the generation stage.
        """
        if isinstance(value, float):
            self._value = value
        elif isinstance(value, int):
            self._value = float(value)
        elif isinstance(value, str) and value in states:
            self._value = states[value][1]
        else:
            log.error(
                f"Unrecognised chunk state {value}. Defaulting to fully generated.\nIf this is a new version report it to the developers."
            )
            self._value = 2.0

    def as_type(self, status_format: StatusFormats) -> Union[int, float, str]:
        """
        Get the generation state in a given format.

        >>> chunk.status.as_type(StatusFormats.Raw)
        2.0
        >>> chunk.status.as_type(StatusFormats.Bedrock)
        2
        >>> chunk.status.as_type(StatusFormats.Java_14)
        'full'

        :param status_format: The format the status should be returned in. Must be from :class:`StatusFormats`.
        :return: The status code in the requested format.
        """
        if not isinstance(status_format, StatusFormats):
            raise ValueError("Version must be one of the states of StatusFormats")

        if status_format == StatusFormats.Raw:
            return self._value
        elif status_format == StatusFormats.Bedrock:  # Bedrock (0, 1 or 2)
            return int(max(min(2, self._value), 0))
        elif status_format in versions:  # Java 1.13/1.14
            value = next(
                (v for v in versions[status_format] if v[0] <= self._value), None
            )
            if value is None:
                value = next((v for v in versions[status_format] if v[0] <= 2.0), None)
            return value[1]
        else:
            raise ValueError(f"Unsupported status format {status_format}")



from __future__ import annotations

import os
from typing import Dict, Iterable
import re
import threading

from amulet_nbt import NamedTag

from amulet.utils import world_utils
from amulet.api.errors import ChunkDoesNotExist
from amulet.api.data_types import (
    ChunkCoordinates,
    RegionCoordinates,
)
from .region import AnvilRegionInterface

InternalDimension = str


ChunkDataType = Dict[str, NamedTag]


class AnvilDimensionManager:
    """
    A class to manage the data for a dimension.
    This can consist of multiple layers. Eg the region layer which contains chunk data and the entities layer which contains entities.
    """

    level_regex = re.compile(r"DIM(?P<level>-?\d+)")

    def __init__(self, directory: str, *, mcc=False, layers=("region",)):
        self._directory = directory
        self._mcc = mcc
        self.__layers: Dict[str, AnvilRegionManager] = {
            layer: AnvilRegionManager(
                os.path.join(self._directory, layer), mcc=self._mcc
            )
            for layer in layers
        }
        self.__default_layer = self.__layers[layers[0]]

    def all_chunk_coords(self) -> Iterable[ChunkCoordinates]:
        yield from self.__default_layer.all_chunk_coords()

    def has_chunk(self, cx: int, cz: int) -> bool:
        return self.__default_layer.has_chunk(cx, cz)

    def unload(self):
        for layer in self.__layers.values():
            layer.unload()

    def get_chunk_data(self, cx: int, cz: int) -> NamedTag:
        """
        Get a NamedTag of a chunk from the database.
        Will raise ChunkDoesNotExist if the region or chunk does not exist
        """
        # get the region key
        return self.__default_layer.get_chunk_data(cx, cz)

    def get_chunk_data_layers(self, cx: int, cz: int) -> ChunkDataType:
        """Get the chunk data for each layer"""
        chunk_data = {}
        for layer_name, layer in self.__layers.items():
            try:
                chunk_data[layer_name] = layer.get_chunk_data(cx, cz)
            except ChunkDoesNotExist:
                pass

        if chunk_data:
            return chunk_data
        else:
            raise ChunkDoesNotExist

    def put_chunk_data(self, cx: int, cz: int, data: NamedTag):
        """pass data to the region file class"""
        self.__default_layer.put_chunk_data(cx, cz, data)

    def put_chunk_data_layers(self, cx: int, cz: int, data_layers: ChunkDataType):
        """Put one or more layers of data"""
        for layer_name, data in data_layers.items():
            if (
                layer_name not in self.__layers
                and layer_name.isalpha()
                and layer_name.islower()
            ):
                self.__layers[layer_name] = AnvilRegionManager(
                    os.path.join(self._directory, layer_name), mcc=self._mcc
                )
            if layer_name in self.__layers:
                self.__layers[layer_name].put_chunk_data(cx, cz, data)

    def delete_chunk(self, cx: int, cz: int):
        for layer in self.__layers.values():
            layer.delete_chunk(cx, cz)


class AnvilRegionManager:
    """A class to manage a directory of region files."""

    def __init__(self, directory: str, *, mcc=False):
        self._directory = directory
        self._regions: Dict[RegionCoordinates, AnvilRegionInterface] = {}
        self._mcc = mcc
        self._lock = threading.RLock()

    def unload(self):
        with self._lock:
            self._regions.clear()

    def _region_path(self, rx, rz) -> str:
        """Get the file path for a region file."""
        return os.path.join(self._directory, f"r.{rx}.{rz}.mca")

    def _has_region(self, rx: int, rz: int) -> bool:
        """Does a region file exist."""
        return os.path.isfile(self._region_path(rx, rz))

    def _get_region(self, rx: int, rz: int, create=False) -> AnvilRegionInterface:
        with self._lock:
            if (rx, rz) in self._regions:
                return self._regions[(rx, rz)]
            elif create or self._has_region(rx, rz):
                region = self._regions[(rx, rz)] = AnvilRegionInterface(
                    self._region_path(rx, rz), mcc=self._mcc
                )
                return region
            else:
                raise ChunkDoesNotExist

    def _iter_regions(self) -> Iterable[AnvilRegionInterface]:
        if os.path.isdir(self._directory):
            for region_file_name in os.listdir(self._directory):
                rx, rz = AnvilRegionInterface.get_coords(region_file_name)
                if rx is None:
                    continue
                yield self._get_region(rx, rz)

    def all_chunk_coords(self) -> Iterable[ChunkCoordinates]:
        for region in self._iter_regions():
            yield from region.all_chunk_coords()

    def has_chunk(self, cx: int, cz: int) -> bool:
        try:
            region = self._get_region(
                *world_utils.chunk_coords_to_region_coords(cx, cz)
            )
        except ChunkDoesNotExist:
            return False
        else:
            return region.has_chunk(cx & 0x1F, cz & 0x1F)

    def get_chunk_data(self, cx: int, cz: int) -> NamedTag:
        """
        Get a NamedTag of a chunk from the database.
        Will raise ChunkDoesNotExist if the region or chunk does not exist
        """
        # get the region key
        return self._get_region(
            *world_utils.chunk_coords_to_region_coords(cx, cz)
        ).get_data(cx & 0x1F, cz & 0x1F)

    def put_chunk_data(self, cx: int, cz: int, data: NamedTag):
        """pass data to the region file class"""
        self._get_region(
            *world_utils.chunk_coords_to_region_coords(cx, cz), create=True
        ).write_data(cx & 0x1F, cz & 0x1F, data)

    def delete_chunk(self, cx: int, cz: int):
        try:
            region = self._get_region(
                *world_utils.chunk_coords_to_region_coords(cx, cz)
            )
        except ChunkDoesNotExist:
            pass
        else:
            region.delete_data(cx & 0x1F, cz & 0x1F)



from __future__ import annotations

import os
import struct
import warnings
import zlib
import gzip
from typing import Tuple, Dict, Union, Optional, BinaryIO, Generator
import numpy
import time
import re
import threading
import logging
from enum import IntEnum

from amulet_nbt import NamedTag, load as load_nbt

from amulet.api.errors import ChunkDoesNotExist, ChunkLoadError
from amulet.api.data_types import (
    ChunkCoordinates,
)
from ._sector_manager import SectorManager, Sector

Depreciated = object()

SectorSize = 0x1000
MaxRegionSize = 255 * SectorSize  # the maximum size data in the region file can be

log = logging.getLogger(__name__)


class RegionFileVersion(IntEnum):
    VERSION_GZIP = 1
    VERSION_DEFLATE = 2
    VERSION_NONE = 3


def _validate_region_coords(cx: int, cz: int):
    """Make sure that the coordinates are in region space."""
    if not (0 <= cx <= 32 and 0 <= cz <= 32):
        raise ValueError("coordinates must be in region space")


def _compress(data: NamedTag) -> bytes:
    """Convert an NBTFile into a compressed bytes object"""
    data = data.save_to(compressed=False)
    return b"\x02" + zlib.compress(data)


def _decompress(data: bytes) -> NamedTag:
    """Convert a bytes object into an NBTFile"""
    compress_type, data = data[0], data[1:]
    if compress_type == RegionFileVersion.VERSION_GZIP:
        return load_nbt(gzip.decompress(data), compressed=False)
    elif compress_type == RegionFileVersion.VERSION_DEFLATE:
        return load_nbt(zlib.decompress(data), compressed=False)
    elif compress_type == RegionFileVersion.VERSION_NONE:
        return load_nbt(data, compressed=False)
    raise ChunkLoadError(f"Invalid compression type {compress_type}")


def _sanitise_file(handler: BinaryIO):
    handler.seek(0, os.SEEK_END)
    file_size = handler.tell()
    if file_size & 0xFFF:
        # ensure the file is a multiple of 4096 bytes
        file_size = (file_size | 0xFFF) + 1
        handler.truncate(file_size)

    # if the length of the region file is less than 8KiB extend it to 8KiB
    if file_size < SectorSize * 2:
        file_size = SectorSize * 2
        handler.truncate(file_size)


class AnvilRegionInterface:
    """
    A class to read and write Minecraft Java Edition Region files.
    Only one class should exist per region file at any given time otherwise bad things may happen.
    """

    region_regex = re.compile(r"r\.(?P<rx>-?\d+)\.(?P<rz>-?\d+)\.mca")

    __slots__ = (
        "_path",
        "_rx",
        "_rz",
        "_mcc",
        "_sector_manager",
        "_chunk_locations",
        "_lock",
    )

    # The path to the region file
    _path: str

    # The region coordinates
    _rx: int
    _rz: int

    # Is support for .mcc files enabled
    _mcc: bool

    # A class to track which sectors are reserved
    _sector_manager: Optional[SectorManager]

    # A dictionary mapping the chunk coordinate to the location on disk
    # Key is a Sector if stored in the region file and str to the file path if stored externally
    _chunk_locations: Dict[ChunkCoordinates, Sector]

    # A lock to limit access to multiple threads
    _lock: threading.RLock

    @staticmethod
    def get_coords(file_path: str) -> Union[Tuple[None, None], Tuple[int, int]]:
        """Parse a region file path to get the region coordinates."""
        file_path = os.path.basename(file_path)
        match = AnvilRegion.region_regex.fullmatch(file_path)
        if match is None:
            return None, None
        return int(match.group("rx")), int(match.group("rz"))

    def __init__(self, file_path: str, create=Depreciated, mcc=False):
        """
        A class wrapper for a region file
        :param file_path: The file path of the region file
        :param create: bool - if true will create the region from scratch. If false will try loading from disk
        """
        self._path = file_path
        self._rx, self._rz = self.get_coords(file_path)
        self._mcc = mcc  # create mcc file if the chunk is greater than 1MiB
        self._sector_manager = None
        self._chunk_locations = {}
        self._lock = threading.RLock()

        if create is not Depreciated:
            warnings.warn(
                "create argument is depreciated. The region will always be created upon writing if it does not exist."
            )

    @property
    def path(self) -> str:
        """The file path to the region file."""
        return self._path

    @property
    def rx(self) -> int:
        """The region x coordinate."""
        return self._rx

    @property
    def rz(self) -> int:
        """The region z coordinate."""
        return self._rz

    def get_mcc_path(self, cx: int, cz: int):
        """Get the mcc path. Coordinates are global chunk coordinates."""
        return os.path.join(
            os.path.dirname(self._path),
            f"c.{cx + self._rx * 32}.{cz + self._rz * 32}.mcc",
        )

    def _load(self):
        with self._lock:
            if self._sector_manager is not None:
                return

            # Create the sector manager and ensure the header is not reservable
            self._sector_manager = SectorManager(0, 0x2000)
            self._sector_manager.reserve(Sector(0, 0x2000))

            if os.path.isfile(self._path):
                # Load the file and populate the sector manager
                with open(self._path, "rb+") as handler:
                    _sanitise_file(handler)
                    handler.seek(0)
                    location_table = numpy.fromfile(
                        handler, dtype=">u4", count=1024
                    ).reshape(32, 32)
                    for (z, x), sector_data in numpy.ndenumerate(location_table):
                        if sector_data:
                            sector_offset = (sector_data >> 8) * 0x1000
                            sector_size = (sector_data & 0xFF) * 0x1000
                            sector = Sector(sector_offset, sector_offset + sector_size)
                            self._sector_manager.reserve(sector)
                            self._chunk_locations[(x, z)] = sector

    def all_chunk_coords(self) -> Generator[ChunkCoordinates, None, None]:
        """An iterable of chunk coordinates in world space."""
        self._load()
        for cx, cz in list(self._chunk_locations):
            yield cx + self.rx * 32, cz + self.rz * 32

    def has_chunk(self, cx: int, cz: int) -> bool:
        """Does the chunk exists. Coords are in region space."""
        _validate_region_coords(cx, cz)
        self._load()
        return (cx, cz) in self._chunk_locations

    def unload(self):
        """Unload the data if it is not being used."""
        with self._lock:
            self._sector_manager = None
            self._chunk_locations.clear()

    def get_data(self, cx: int, cz: int) -> NamedTag:
        _validate_region_coords(cx, cz)
        self._load()
        sector = self._chunk_locations.get((cx, cz))
        if sector is None:
            raise ChunkDoesNotExist
        with self._lock:
            os.makedirs(os.path.dirname(self._path), exist_ok=True)
            with open(self._path, "rb+") as handler:
                _sanitise_file(handler)
                handler.seek(0, os.SEEK_END)
                if handler.tell() < sector.stop:
                    # if the sector is beyond the end of the file
                    raise ChunkDoesNotExist

                handler.seek(sector.start)
                buffer_size_bytes: bytes = handler.read(4)
                buffer_size = struct.unpack(">I", buffer_size_bytes)[0]
                buffer: bytes = handler.read(buffer_size)

                if buffer:
                    if buffer[0] & 128:  # if the "external" bit is set
                        if self._mcc:
                            mcc_path = self.get_mcc_path(cx, cz)
                            if os.path.isfile(mcc_path):
                                with open(mcc_path, "rb") as f:
                                    return _decompress(
                                        bytes([buffer[0] & 127]) + f.read()
                                    )
                    else:
                        return _decompress(buffer)
        raise ChunkDoesNotExist

    def _write_data(self, cx: int, cz: int, data: Optional[bytes]):
        _validate_region_coords(cx, cz)
        if isinstance(data, bytes) and len(data) + 4 > MaxRegionSize and not self._mcc:
            # if the data is too large and mcc files are not supported then do nothing
            log.error(
                f"Could not save data {cx},{cz} in region file {self._path} because it was too large."
            )
            return

        self._load()
        with self._lock:
            os.makedirs(os.path.dirname(self._path), exist_ok=True)
            with open(
                self._path, "rb+" if os.path.isfile(self._path) else "wb+"
            ) as handler:
                handler: BinaryIO
                _sanitise_file(handler)

                old_sector = self._chunk_locations.pop((cx, cz), None)
                if old_sector is not None:
                    # the chunk used to exist
                    handler.seek(old_sector.start + 4)
                    if self._mcc and handler.read(1)[0] & 127:
                        # if the file is stored externally delete the file
                        mcc_path = self.get_mcc_path(cx, cz)
                        if os.path.isfile(mcc_path):
                            os.remove(mcc_path)
                    self._sector_manager.free(old_sector)

                location = b"\x00\x00\x00\x00"

                if isinstance(data, bytes):
                    # find a memory location large enough to fit the data
                    if len(data) + 4 > MaxRegionSize:
                        # save externally (if mcc files are not supported the check at the top will filter large files out)
                        with open(self.get_mcc_path(cx, cz), "wb") as mcc:
                            mcc.write(data[1:])
                        data = bytes([data[0] | 128])
                    data = struct.pack(">I", len(data)) + data
                    sector_length = (len(data) | 0xFFF) + 1
                    sector = self._sector_manager.reserve_space(sector_length)
                    assert sector.start & 0xFFF == 0
                    self._chunk_locations[(cx, cz)] = sector
                    location = struct.pack(
                        ">I", (sector.start >> 4) + (sector_length >> 12)
                    )
                    handler.seek(sector.start)
                    handler.write(data)
                    _sanitise_file(handler)

                # write the header data
                handler.seek(4 * (cx + cz * 32))
                handler.write(location)
                handler.seek(SectorSize - 4, os.SEEK_CUR)
                handler.write(struct.pack(">I", int(time.time())))

    def write_data(self, cx: int, cz: int, data: NamedTag):
        """Write the data to the region file."""
        bytes_data = _compress(data)
        self._write_data(cx, cz, bytes_data)

    def delete_data(self, cx: int, cz: int):
        """Delete the data from the region file."""
        self._write_data(cx, cz, None)


class BufferedAnvilRegionInterface(AnvilRegionInterface):
    """An interface to an anvil region file with a buffer before writing."""

    __slots__ = ("_buffer",)

    _buffer: Dict[ChunkCoordinates, Optional[bytes]]

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "BufferedAnvilRegionInterface aka AnvilRegion is depreciated and replaced by AnvilRegionInterface",
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)
        self._buffer = {}

    def all_chunk_coords(self) -> Generator[ChunkCoordinates, None, None]:
        """An iterable of chunk coordinates in world space."""
        self._load()
        with self._lock:
            chunks = [
                (cx + self.rx * 32, cz + self.rz * 32)
                for cx, cz in self._chunk_locations
                if (cx, cz) not in self._buffer
            ] + [
                (cx + self.rx * 32, cz + self.rz * 32)
                for (cx, cz), value in self._buffer.items()
                if value is not None
            ]
        yield from chunks

    def unload(self):
        with self._lock:
            super().unload()
            self._buffer.clear()

    def has_chunk(self, cx: int, cz: int) -> bool:
        """Does the chunk exists. Coords are in region space."""
        _validate_region_coords(cx, cz)
        with self._lock:
            if (cx, cz) in self._buffer:
                return self._buffer[(cx, cz)] is not None
            else:
                self._load()
                return (cx, cz) in self._chunk_locations

    def get_chunk_data(self, cx: int, cz: int) -> NamedTag:
        """Get chunk data. Coords are in region space."""
        _validate_region_coords(cx, cz)
        data = self._buffer.get((cx, cz))
        if data is not None:
            return _decompress(data)
        return self.get_data(cx, cz)

    def put_chunk_data(self, cx: int, cz: int, data: NamedTag):
        """
        Put data to be added to the region file. `save` will push the changes to disk.
        Coords are in region space.
        """
        _validate_region_coords(cx, cz)
        with self._lock:
            self._buffer[(cx, cz)] = _compress(data)

    def delete_chunk_data(self, cx: int, cz: int):
        """
        Mark the data for deletion. `save` will push the changes to disk.
        Coords are in region space.
        """
        _validate_region_coords(cx, cz)
        with self._lock:
            self._buffer[(cx, cz)] = None

    def save(self):
        """Write the buffered data to the file and clear the buffer."""
        with self._lock:
            for (cx, cz), data in self._buffer.items():
                self._write_data(cx, cz, data)
            self._buffer.clear()


AnvilRegion = BufferedAnvilRegionInterface



from __future__ import annotations

from .anvil_2844 import (
    Anvil2844Interface as ParentInterface,
)
from amulet.api.chunk import StatusFormats


class Anvil3463Interface(ParentInterface):
    def __init__(self):
        super().__init__()
        self._set_feature("status", StatusFormats.Java_20)

    @staticmethod
    def minor_is_valid(key: int):
        return 3454 <= key < 3480


export = Anvil3463Interface



import os
from typing import (
    Optional,
    Tuple,
    Iterable,
    TYPE_CHECKING,
    BinaryIO,
    Dict,
    Union,
    List,
    NamedTuple,
)
import numpy
import copy

from amulet_nbt import (
    ShortTag,
    StringTag,
    ListTag,
    CompoundTag,
    ByteArrayTag,
    NamedTag,
    load as load_nbt,
)

from amulet.api.data_types import (
    AnyNDArray,
    VersionNumberAny,
    VersionNumberTuple,
    Dimension,
    PlatformType,
    PointCoordinates,
    ChunkCoordinates,
)
from amulet.api.registry import BlockManager
from amulet.api.wrapper import StructureFormatWrapper
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionBox, SelectionGroup
from amulet.api.errors import ObjectReadError, ObjectWriteError, ChunkDoesNotExist
from .interface import (
    JavaSchematicInterface,
    BedrockSchematicInterface,
    SchematicInterface,
)
from .chunk import SchematicChunk

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator, Interface

java_interface = JavaSchematicInterface()
bedrock_interface = BedrockSchematicInterface()


def _is_schematic(path: str):
    """Check if a file is actually a sponge schematic file."""
    try:
        return "Blocks" in load_nbt(path).compound
    except:
        return False


class SchematicFormatWrapper(StructureFormatWrapper[VersionNumberTuple]):
    """
    This FormatWrapper class exists to interface with the legacy schematic structure format.
    """

    def __init__(self, path: str):
        """
        Construct a new instance of :class:`SchematicFormatWrapper`.

        This should not be used directly. You should instead use :func:`amulet.load_format`.

        :param path: The file path to the serialised data.
        """
        super().__init__(path)
        self._chunks: Dict[
            ChunkCoordinates,
            SchematicChunk,
        ] = {}

    def _create(
        self,
        overwrite: bool,
        bounds: Union[
            SelectionGroup, Dict[Dimension, Optional[SelectionGroup]], None
        ] = None,
        **kwargs,
    ):
        if not overwrite and os.path.isfile(self.path):
            raise ObjectWriteError(f"There is already a file at {self.path}")
        if self._platform == "bedrock":
            self._version = (1, 12, 0)
        else:
            self._platform = "java"
            self._version = (1, 12, 2)
        self._chunks = {}
        self._set_selection(bounds)
        self._is_open = True
        self._has_lock = True

    def open_from(self, f: BinaryIO):
        schematic = load_nbt(f).compound
        if "BlockData" in schematic:
            raise ObjectReadError("This file is not a legacy schematic file.")
        materials = schematic.get_string("Materials", StringTag("Alpha")).py_str
        if materials == "Alpha":
            self._platform = "java"
            self._version = (1, 12, 2)
        elif materials == "Pocket":
            self._platform = "bedrock"
            self._version = (1, 12, 0)
        else:
            raise Exception(
                f'"{materials}" is not a supported platform for a schematic file.'
            )
        self._chunks = {}
        selection_box = SelectionBox(
            (0, 0, 0),
            (
                schematic.get_short("Width").py_int,
                schematic.get_short("Height").py_int,
                schematic.get_short("Length").py_int,
            ),
        )
        self._bounds[self.dimensions[0]] = SelectionGroup(selection_box)
        entities: ListTag = schematic.get_list("Entities", ListTag())
        block_entities: ListTag = schematic.get_list("TileEntities", ListTag())
        blocks: numpy.ndarray = (
            schematic.get_byte_array("Blocks")
            .np_array.astype(numpy.uint8)
            .astype(numpy.uint16)
        )
        if "AddBlocks" in schematic:
            add_blocks = schematic.get_byte_array("AddBlocks").np_array
            blocks = (
                blocks
                + (
                    numpy.concatenate([[(add_blocks & 0xF0) >> 4], [add_blocks & 0xF]])
                    .T.ravel()
                    .astype(numpy.uint16)
                    << 8
                )[: blocks.size]
            )
        max_point = selection_box.max
        temp_shape = (max_point[1], max_point[2], max_point[0])
        blocks = numpy.transpose(blocks.reshape(temp_shape), (2, 0, 1))  # YZX => XYZ
        data = numpy.transpose(
            schematic.get_byte_array("Data").np_array.reshape(temp_shape), (2, 0, 1)
        ).astype(numpy.uint8)
        for cx, cz in selection_box.chunk_locations():
            box = SelectionBox(
                (cx * self.sub_chunk_size, 0, cz * self.sub_chunk_size),
                (
                    min((cx + 1) * self.sub_chunk_size, selection_box.size_x),
                    selection_box.size_y,
                    min((cz + 1) * self.sub_chunk_size, selection_box.size_z),
                ),
            )
            self._chunks[(cx, cz)] = SchematicChunk(
                box, blocks[box.slice], data[box.slice], [], []
            )
        for e in block_entities:
            if isinstance(e, CompoundTag) and all(key in e for key in ("x", "y", "z")):
                x = e.get_int("x").py_int
                y = e.get_int("y").py_int
                z = e.get_int("z").py_int
                if (x, y, z) in selection_box:
                    cx = x >> 4
                    cz = z >> 4
                    self._chunks[(cx, cz)].block_entities.append(e)
        for e in entities:
            if isinstance(e, CompoundTag) and "Pos" in e:
                pos: PointCoordinates = tuple(map(float, e.get_list("Pos", ListTag())))
                if len(pos) == 3:
                    if pos in selection_box:
                        cx = int(pos[0]) >> 4
                        cz = int(pos[2]) >> 4
                        self._chunks[(cx, cz)].entities.append(e)

    @staticmethod
    def is_valid(path: str) -> bool:
        return (
            os.path.isfile(path) and path.endswith(".schematic") and _is_schematic(path)
        )

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {"java": (True, True), "bedrock": (True, True)}

    @property
    def extensions(self) -> Tuple[str, ...]:
        return (".schematic",)

    def _get_interface(self, raw_chunk_data=None) -> SchematicInterface:
        if self._platform == "java":
            return java_interface
        elif self._platform == "bedrock":
            return bedrock_interface
        else:
            raise Exception(f"{self._platform} is not a supported platform")

    def _get_interface_and_translator(
        self, raw_chunk_data=None
    ) -> Tuple["Interface", "Translator", VersionNumberAny]:
        interface = self._get_interface(raw_chunk_data)
        translator, version_identifier = interface.get_translator(
            self.max_world_version, raw_chunk_data, self.translation_manager
        )
        return interface, translator, version_identifier

    def save_to(self, f: BinaryIO):
        if self._platform == "java":
            materials = "Alpha"
        elif self._platform == "bedrock":
            materials = "Pocket"
        else:
            raise ObjectWriteError(
                f'"{self._platform}" is not a supported platform for a schematic file.'
            )

        selection = self._bounds[self.dimensions[0]].selection_boxes[0]

        tag = CompoundTag(
            {
                "TileTicks": ListTag(),
                "Width": ShortTag(selection.size_x),
                "Height": ShortTag(selection.size_y),
                "Length": ShortTag(selection.size_z),
                "Materials": StringTag(materials),
            }
        )

        entities = []
        block_entities = []
        blocks = numpy.zeros(
            selection.shape, dtype=numpy.uint16
        )  # only 12 bits are actually used at most
        block_data = numpy.zeros(
            selection.shape, dtype=numpy.uint8
        )  # only 4 bits are used

        for chunk in self._chunks.values():
            if chunk.selection.intersects(selection):
                box = chunk.selection.create_moved_box(selection.min, subtract=True)
                blocks[box.slice] = chunk.blocks
                block_data[box.slice] = chunk.data
                for be in chunk.block_entities:
                    coord_type = be["x"].__class__
                    be["x"] = coord_type(be["x"].py_int - selection.min_x)
                    be["y"] = coord_type(be["y"].py_int - selection.min_y)
                    be["z"] = coord_type(be["z"].py_int - selection.min_z)
                    block_entities.append(be)
                for e in chunk.entities:
                    coord_type = e["Pos"][0].__class__
                    e["Pos"][0] = coord_type(e["Pos"][0] - selection.min_x)
                    e["Pos"][1] = coord_type(e["Pos"][1] - selection.min_y)
                    e["Pos"][2] = coord_type(e["Pos"][2] - selection.min_z)
                    entities.append(e)

        tag["Entities"] = ListTag(entities)
        tag["TileEntities"] = ListTag(block_entities)
        tag["Data"] = ByteArrayTag(numpy.transpose(block_data, (1, 2, 0)))  # XYZ => YZX
        tag["Blocks"] = ByteArrayTag(
            numpy.transpose((blocks & 0xFF).astype(numpy.uint8), (1, 2, 0))
        )
        if numpy.max(blocks) > 0xFF:
            add_blocks = (numpy.transpose(blocks & 0xF00, (1, 2, 0)) >> 8).ravel()
            tag["AddBlocks"] = ByteArrayTag((add_blocks[::2] << 4) + add_blocks[1::2])
        NamedTag(tag, "Schematic").save_to(f)

    def _close(self):
        """Close the disk database"""
        self._chunks.clear()

    def unload(self):
        pass

    def all_chunk_coords(
        self, dimension: Optional[Dimension] = None
    ) -> Iterable[ChunkCoordinates]:
        yield from self._chunks.keys()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return (cx, cz) in self._chunks

    def _pack(
        self,
        chunk: "Chunk",
        translator: "Translator",
        chunk_version: VersionNumberAny,
    ) -> Tuple["Chunk", AnyNDArray]:
        version = self.translation_manager.get_version(
            *translator.translator_key(chunk_version)
        )
        palette = []
        for entry in chunk.block_palette.blocks:
            b = version.block.block_to_ints(entry)
            if b is None:
                b = (0, 0)
            palette.append(b)
        return (
            chunk,
            numpy.array(palette),
        )

    def _encode(
        self,
        interface: SchematicInterface,
        chunk: Chunk,
        dimension: Dimension,
        chunk_palette: AnyNDArray,
    ):
        return interface.encode(
            chunk,
            chunk_palette,
            self.max_world_version,
            SelectionBox.create_chunk_box(chunk.cx, chunk.cz).intersection(
                self._bounds[dimension].to_box()
            ),
        )

    def _unpack(
        self,
        translator: "Translator",
        game_version: VersionNumberAny,
        chunk: "Chunk",
        chunk_palette: AnyNDArray,
    ) -> "Chunk":
        version = self.translation_manager.get_version(
            *translator.translator_key(game_version)
        )
        palette = chunk._block_palette = BlockManager()
        lut = numpy.array(
            [
                palette.get_add_block(version.block.ints_to_block(block, data))
                for block, data in chunk_palette
            ]
        )
        if len(palette.blocks) != len(chunk_palette):
            # if a blockstate was defined twice
            for cy in chunk.blocks.sub_chunks:
                chunk.blocks.add_sub_chunk(cy, lut[chunk.blocks.get_sub_chunk(cy)])

        return chunk

    def _delete_chunk(self, cx: int, cz: int, dimension: Optional[Dimension] = None):
        if (cx, cz) in self._chunks:
            del self._chunks[(cx, cz)]

    def _put_raw_chunk_data(
        self,
        cx: int,
        cz: int,
        section: SchematicChunk,
        dimension: Optional[Dimension] = None,
    ):
        self._chunks[(cx, cz)] = copy.deepcopy(section)

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: Optional[Dimension] = None
    ) -> SchematicChunk:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        if (cx, cz) in self._chunks:
            return copy.deepcopy(self._chunks[(cx, cz)])
        else:
            raise ChunkDoesNotExist



from __future__ import annotations

from .anvil_1901 import Anvil1901Interface as ParentInterface


class Anvil1908Interface(ParentInterface):
    """
    Changed height keys again again
    """

    def __init__(self):
        super().__init__()
        self._set_feature("height_map", "C|V4")

    @staticmethod
    def minor_is_valid(key: int):
        return 1908 <= key < 1912


export = Anvil1908Interface



from __future__ import annotations

from .base_level import BaseLevel
from amulet.api.wrapper import WorldFormatWrapper


class World(BaseLevel):
    """
    Class that handles world editing of any world format via an separate and flexible data format
    """

    def __init__(self, directory: str, world_wrapper: WorldFormatWrapper):
        assert isinstance(world_wrapper, WorldFormatWrapper)
        super().__init__(directory, world_wrapper)

    @property
    def level_wrapper(self) -> "WorldFormatWrapper":
        """A class to access data directly from the level."""
        return self._level_wrapper



from __future__ import annotations

from typing import Tuple, Optional, TYPE_CHECKING
import copy
import math
import logging

import numpy

from amulet import entity_support
from amulet.api.registry import BlockManager, BiomeManager
from amulet.api.block import Block
from amulet.api.block_entity import BlockEntity
from amulet.api.entity import Entity
from amulet.api.chunk import Chunk, BiomesShape
from amulet.api.data_types import (
    AnyNDArray,
    BlockNDArray,
    BlockCoordinates,
    GetChunkCallback,
    TranslateBlockCallback,
    TranslateEntityCallback,
    GetBlockCallback,
    TranslateBlockCallbackReturn,
    TranslateEntityCallbackReturn,
    VersionNumberAny,
    VersionIdentifierType,
)

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from PyMCTranslate import Version, TranslationManager


class Translator:
    def translator_key(self, version_number: VersionNumberAny) -> VersionIdentifierType:
        return self._translator_key(version_number)

    def _translator_key(
        self, version_number: VersionNumberAny
    ) -> VersionIdentifierType:
        """
        Return the version key for PyMCTranslate

        :return: The tuple version key for PyMCTranslate
        """
        raise NotImplementedError

    @staticmethod
    def is_valid(key: Tuple) -> bool:
        """
        Returns whether this translator is able to translate the chunk type with a given identifier key,
        generated by the decoder.

        :param key: The key who's decodability needs to be checked.
        :return: True if this translator is able to translate the chunk type associated with the key, False otherwise.
        """
        raise NotImplementedError

    @staticmethod
    def _translate(
        chunk: Chunk,
        get_chunk_callback: Optional[GetChunkCallback],
        translate_block: TranslateBlockCallback,
        translate_entity: TranslateEntityCallback,
        full_translate: bool,
    ):
        if full_translate:
            todo = []
            output_block_entities = []
            output_entities = []
            finished = BlockManager()
            palette_mappings = {}

            # translate each block without using the callback
            for i, input_block in enumerate(chunk.block_palette):
                input_block: Block
                (
                    output_block,
                    output_block_entity,
                    output_entity,
                    extra,
                ) = translate_block(input_block, None, (0, 0, 0))
                if extra and get_chunk_callback:
                    todo.append(i)
                elif output_block is not None:
                    palette_mappings[i] = finished.get_add_block(output_block)
                    if output_block_entity is not None:
                        for cy in chunk.blocks.sub_chunks:
                            for x, y, z in zip(
                                *numpy.where(chunk.blocks.get_sub_chunk(cy) == i)
                            ):
                                output_block_entities.append(
                                    output_block_entity.new_at_location(
                                        x + chunk.cx * 16,
                                        y + cy * 16,
                                        z + chunk.cz * 16,
                                    )
                                )
                else:
                    # TODO: this should only happen if the object is an entity, set the block to air
                    pass

                if output_entity and entity_support:
                    for cy in chunk.blocks.sub_chunks:
                        for x, y, z in zip(
                            *numpy.where(chunk.blocks.get_sub_chunk(cy) == i)
                        ):
                            x += chunk.cx * 16
                            y += cy * 16
                            z += chunk.cz * 16
                            for entity in output_entity:
                                e = copy.deepcopy(entity)
                                e.location += (x, y, z)
                                output_entities.append(e)

            # re-translate the blocks that require extra information
            block_mappings = {}
            for index in todo:
                for cy in chunk.blocks.sub_chunks:
                    for x, y, z in zip(
                        *numpy.where(chunk.blocks.get_sub_chunk(cy) == index)
                    ):
                        y += cy * 16

                        def get_block_at(
                            pos: BlockCoordinates,
                        ) -> Tuple[Block, Optional[BlockEntity]]:
                            """Get a block at a location relative to the current block"""
                            nonlocal x, y, z, chunk, cy

                            # calculate position relative to chunk base
                            dx, dy, dz = pos
                            dx += x
                            dy += y
                            dz += z

                            abs_x = dx + chunk.cx * 16
                            abs_y = dy
                            abs_z = dz + chunk.cz * 16

                            # calculate relative chunk position
                            cx = dx // 16
                            cz = dz // 16
                            if cx == 0 and cz == 0:
                                # if it is the current chunk
                                block = chunk.block_palette[chunk.blocks[dx, dy, dz]]
                                return (
                                    block,
                                    chunk.block_entities.get((abs_x, abs_y, abs_z)),
                                )

                            # if it is in a different chunk
                            local_chunk = get_chunk_callback(cx, cz)
                            block = local_chunk.block_palette[
                                local_chunk.blocks[dx % 16, dy, dz % 16]
                            ]
                            return (
                                block,
                                local_chunk.block_entities.get((abs_x, abs_y, abs_z)),
                            )

                        input_block = chunk.block_palette[chunk.blocks[x, y, z]]
                        (
                            output_block,
                            output_block_entity,
                            output_entity,
                            _,
                        ) = translate_block(
                            input_block,
                            get_block_at,
                            (x + chunk.cx * 16, y, z + chunk.cz * 16),
                        )
                        if output_block is not None:
                            block_mappings[(x, y, z)] = finished.get_add_block(
                                output_block
                            )
                            if output_block_entity is not None:
                                output_block_entities.append(
                                    output_block_entity.new_at_location(
                                        x + chunk.cx * 16, y, z + chunk.cz * 16
                                    )
                                )
                        else:
                            # TODO: set the block to air
                            pass

                        if output_entity and entity_support:
                            for entity in output_entity:
                                e = copy.deepcopy(entity)
                                e.location += (x, y, z)
                                output_entities.append(e)

            if entity_support:
                for entity in chunk.entities:
                    output_block, output_block_entity, output_entity = translate_entity(
                        entity
                    )
                    if output_block is not None:
                        block_location = (
                            int(math.floor(entity.x)),
                            int(math.floor(entity.y)),
                            int(math.floor(entity.z)),
                        )
                        block_mappings[block_location] = output_block
                        if output_block_entity:
                            output_block_entities.append(
                                output_block_entity.new_at_location(*block_location)
                            )
                    if output_entity:
                        for e in output_entity:
                            e.location = entity.location
                            output_entities.append(e)

            for cy in chunk.blocks.sub_chunks:
                old_blocks = chunk.blocks.get_sub_chunk(cy)
                new_blocks = numpy.zeros(old_blocks.shape, dtype=old_blocks.dtype)
                for old, new in palette_mappings.items():
                    new_blocks[old_blocks == old] = new
                chunk.blocks.add_sub_chunk(cy, new_blocks)
            for (x, y, z), new in block_mappings.items():
                chunk.blocks[x, y, z] = new
            chunk.block_entities = output_block_entities
            chunk.entities = output_entities
            chunk._block_palette = finished

    def to_universal(
        self,
        chunk_version: VersionNumberAny,
        translation_manager: "TranslationManager",
        chunk: Chunk,
        get_chunk_callback: Optional[GetChunkCallback],
        full_translate: bool,
    ) -> Chunk:
        """
        Translate an interface-specific chunk into the universal format.

        :param chunk_version: The version number (int or tuple) of the input chunk
        :param translation_manager: TranslationManager used for the translation
        :param chunk: The chunk to translate.
        :param get_chunk_callback: function callback to get a chunk's data
        :param full_translate: if true do a full translate. If false just unpack the block_palette (used in callback)
        :return: Chunk object in the universal format.
        """
        version = translation_manager.get_version(*self._translator_key(chunk_version))
        self._biomes_to_universal(version, chunk)
        self._blocks_entities_to_universal(
            chunk_version,
            translation_manager,
            chunk,
            get_chunk_callback,
            full_translate,
        )
        return chunk

    def _blocks_entities_to_universal(
        self,
        chunk_version: VersionNumberAny,
        translation_manager: "TranslationManager",
        chunk: Chunk,
        get_chunk_callback: Optional[GetChunkCallback],
        full_translate: bool,
    ):
        version = translation_manager.get_version(*self._translator_key(chunk_version))

        def translate_block(
            input_object: Block,
            get_block_callback: Optional[GetBlockCallback],
            block_location: BlockCoordinates,
        ) -> TranslateBlockCallbackReturn:
            final_block = None
            final_block_entity = None
            final_entities = []
            final_extra = False

            for depth, block in enumerate(input_object.block_tuple):
                (
                    output_object,
                    output_block_entity,
                    extra,
                ) = version.block.to_universal(
                    block,
                    get_block_callback=get_block_callback,
                    block_location=block_location,
                )

                if isinstance(output_object, Block):
                    if not output_object.namespace.startswith("universal"):
                        log.debug(
                            f"Error translating {input_object.full_blockstate} to universal. Got {output_object.full_blockstate}"
                        )
                    if final_block is None:
                        final_block = output_object
                    else:
                        final_block += output_object
                    if depth == 0:
                        final_block_entity = output_block_entity

                elif isinstance(output_object, Entity):
                    final_entities.append(output_object)
                    # TODO: offset entity coords

                final_extra |= extra

            return final_block, final_block_entity, final_entities, final_extra

        def translate_entity(input_object: Entity) -> TranslateEntityCallbackReturn:
            final_block = None
            final_block_entity = None
            final_entities = []
            # TODO
            return final_block, final_block_entity, final_entities

        self._translate(
            chunk,
            get_chunk_callback,
            translate_block,
            translate_entity,
            full_translate,
        )

    @staticmethod
    def _biomes_to_universal(translator_version: "Version", chunk: Chunk):
        chunk._biome_palette = BiomeManager(
            [
                translator_version.biome.to_universal(biome)
                for biome in chunk.biome_palette
            ]
        )

    def from_universal(
        self,
        max_world_version_number: VersionNumberAny,
        translation_manager: "TranslationManager",
        chunk: Chunk,
        get_chunk_callback: Optional[GetChunkCallback],
        full_translate: bool,
    ) -> Chunk:
        """
        Translate a universal chunk into the interface-specific format.

        :param max_world_version_number: The version number (int or tuple) of the max world version
        :param translation_manager: TranslationManager used for the translation
        :param chunk: The chunk to translate.
        :param get_chunk_callback: function callback to get a chunk's data
        :param full_translate: if true do a full translate. If false just pack the block_palette (used in callback)
        :return: Chunk object in the interface-specific format and block_palette.
        """
        version = translation_manager.get_version(
            *self._translator_key(max_world_version_number)
        )
        self._blocks_entities_from_universal(
            max_world_version_number,
            translation_manager,
            chunk,
            get_chunk_callback,
            full_translate,
        )
        self._biomes_from_universal(version, chunk)
        return chunk

    def _blocks_entities_from_universal(
        self,
        max_world_version_number: VersionNumberAny,
        translation_manager: "TranslationManager",
        chunk: Chunk,
        get_chunk_callback: Optional[GetChunkCallback],
        full_translate: bool,
    ):
        version = translation_manager.get_version(
            *self._translator_key(max_world_version_number)
        )

        # TODO: perhaps find a way so this code isn't duplicated in three places
        def translate_block(
            input_object: Block,
            get_block_callback: Optional[GetBlockCallback],
            block_location: BlockCoordinates,
        ) -> TranslateBlockCallbackReturn:
            final_block = None
            final_block_entity = None
            final_entities = []
            final_extra = False

            for depth, block in enumerate(input_object.block_tuple):
                (
                    output_object,
                    output_block_entity,
                    extra,
                ) = version.block.from_universal(
                    block,
                    get_block_callback=get_block_callback,
                    block_location=block_location,
                )

                if isinstance(output_object, Block):
                    if __debug__ and output_object.namespace.startswith("universal"):
                        log.debug(
                            f"Error translating {input_object.blockstate} from universal. Got {output_object.blockstate}"
                        )
                    if final_block is None:
                        final_block = output_object
                    else:
                        final_block += output_object
                    if depth == 0:
                        final_block_entity = output_block_entity

                elif isinstance(output_object, Entity):
                    final_entities.append(output_object)
                    # TODO: offset entity coords

                final_extra |= extra

            return final_block, final_block_entity, final_entities, final_extra

        def translate_entity(input_object: Entity) -> TranslateEntityCallbackReturn:
            final_block = None
            final_block_entity = None
            final_entities = []
            # TODO
            return final_block, final_block_entity, final_entities

        self._translate(
            chunk,
            get_chunk_callback,
            translate_block,
            translate_entity,
            full_translate,
        )

    @staticmethod
    def _biomes_from_universal(translator_version: "Version", chunk: Chunk):
        chunk._biome_palette = BiomeManager(
            [
                translator_version.biome.from_universal(biome)
                for biome in chunk.biome_palette
            ]
        )

    def unpack(
        self,
        chunk_version: VersionNumberAny,
        translation_manager: "TranslationManager",
        chunk: Chunk,
        palette: AnyNDArray,
    ) -> Chunk:
        """
        Unpack the version-specific block_palette into the stringified version where needed.

        :return: The block_palette converted to block objects.
        """
        version_identifier = self._translator_key(chunk_version)
        self._unpack_blocks(translation_manager, version_identifier, chunk, palette)
        self._unpack_biomes(translation_manager, version_identifier, chunk)
        return chunk

    @staticmethod
    def _unpack_blocks(
        translation_manager: "TranslationManager",
        version_identifier: VersionIdentifierType,
        chunk: Chunk,
        block_palette: AnyNDArray,
    ):
        """
        Unpack the version-specific block_palette into the stringified version where needed.
        :return: The block_palette converted to block objects.
        """
        chunk._block_palette = BlockManager(block_palette)

    @staticmethod
    def _unpack_biomes(
        translation_manager: "TranslationManager",
        version_identifier: VersionIdentifierType,
        chunk: Chunk,
    ):
        """
        Unpack the version-specific biome_palette into the stringified version where needed.
        :return: The biome_palette converted to biome objects.
        """
        version = translation_manager.get_version(*version_identifier)

        if chunk.biomes.dimension == BiomesShape.Shape2D:
            biome_int_palette, biome_array = numpy.unique(
                chunk.biomes, return_inverse=True
            )
            chunk.biomes = biome_array.reshape(chunk.biomes.shape)
            chunk._biome_palette = BiomeManager(
                [version.biome.unpack(biome) for biome in biome_int_palette]
            )
        elif chunk.biomes.dimension == BiomesShape.Shape3D:
            biomes = {}
            palette = []
            palette_length = 0
            for sy in chunk.biomes.sections:
                biome_int_palette, biome_array = numpy.unique(
                    chunk.biomes.get_section(sy), return_inverse=True
                )
                biomes[sy] = (
                    biome_array.reshape(chunk.biomes.section_shape) + palette_length
                )
                palette_length += len(biome_int_palette)
                palette.append(biome_int_palette)

            if palette:
                chunk_palette, lut = numpy.unique(
                    numpy.concatenate(palette), return_inverse=True
                )
                lut = lut.astype(numpy.uint32)
                for sy in biomes:
                    biomes[sy] = lut[biomes[sy]]

                chunk.biomes = biomes
                chunk._biome_palette = BiomeManager(
                    numpy.vectorize(version.biome.unpack)(chunk_palette)
                )

    def pack(
        self,
        max_world_version_number: VersionNumberAny,
        translation_manager: "TranslationManager",
        chunk: Chunk,
    ) -> Tuple[Chunk, AnyNDArray]:
        """
        Translate the list of block objects into a version-specific block_palette.
        :return: The block_palette converted into version-specific blocks (ie id, data tuples for 1.12)
        """
        version_identifier = self._translator_key(max_world_version_number)
        version = translation_manager.get_version(*version_identifier)
        self._pack_biomes(translation_manager, version_identifier, chunk)
        return (
            chunk,
            self._pack_block_palette(version, numpy.array(chunk.block_palette.blocks)),
        )

    def _pack_block_palette(
        self, version: "Version", palette: BlockNDArray
    ) -> AnyNDArray:
        """
        Pack the list of block objects into a version-specific block_palette.
        :return: The block_palette converted into version-specific blocks (ie id, data tuples for 1.12)
        """
        return palette

    @staticmethod
    def _pack_biomes(
        translation_manager: "TranslationManager",
        version_identifier: VersionIdentifierType,
        chunk: Chunk,
    ):
        """
        Unpack the version-specific biome_palette into the stringified version where needed.
        :return: The biome_palette converted to biome objects.
        """
        version = translation_manager.get_version(*version_identifier)

        biome_palette = numpy.array(
            [version.biome.pack(biome) for biome in chunk.biome_palette], numpy.uint32
        )
        if chunk.biomes.dimension == BiomesShape.Shape2D:
            chunk.biomes = biome_palette[chunk.biomes]
        elif chunk.biomes.dimension == BiomesShape.Shape3D:
            chunk.biomes = {
                sy: biome_palette[chunk.biomes.get_section(sy)]
                for sy in chunk.biomes.sections
            }
        chunk._biome_palette = BiomeManager()



import os
from typing import Optional, Tuple, Iterable, TYPE_CHECKING, BinaryIO, Dict, List, Union
import numpy
import copy

from amulet_nbt import (
    IntTag,
    StringTag,
    ListTag,
    CompoundTag,
    load as load_nbt,
    utf8_escape_decoder,
    utf8_escape_encoder,
)

from amulet.api.data_types import (
    VersionNumberAny,
    VersionNumberTuple,
    ChunkCoordinates,
    AnyNDArray,
    Dimension,
    PlatformType,
)
from amulet.api.wrapper import StructureFormatWrapper
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.errors import ChunkDoesNotExist, ObjectWriteError
from amulet.utils.numpy_helpers import brute_sort_objects_no_hash

from .chunk import MCStructureChunk
from .interface import MCStructureInterface

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator, Interface

mcstructure_interface = MCStructureInterface()


class MCStructureFormatWrapper(StructureFormatWrapper[VersionNumberTuple]):
    """
    This FormatWrapper class exists to interface with the Bedrock mcstructure structure block format.
    """

    def __init__(self, path: str):
        """
        Construct a new instance of :class:`MCStructureFormatWrapper`.

        This should not be used directly. You should instead use :func:`amulet.load_format`.

        :param path: The file path to the serialised data.
        """
        super().__init__(path)
        self._chunks: Dict[
            ChunkCoordinates,
            Tuple[
                SelectionBox,
                numpy.ndarray,
                AnyNDArray,
                List[CompoundTag],
                List[CompoundTag],
            ],
        ] = {}

    def _create(
        self,
        overwrite: bool,
        bounds: Union[
            SelectionGroup, Dict[Dimension, Optional[SelectionGroup]], None
        ] = None,
        **kwargs,
    ):
        if not overwrite and os.path.isfile(self.path):
            raise ObjectWriteError(f"There is already a file at {self.path}")
        translator_version = self.translation_manager.get_version(
            "bedrock", (999, 999, 999)
        )
        self._platform = translator_version.platform
        self._version = translator_version.version_number
        self._chunks = {}
        self._set_selection(bounds)
        self._is_open = True
        self._has_lock = True

    def open_from(self, f: BinaryIO):
        mcstructure = load_nbt(
            f, little_endian=True, string_decoder=utf8_escape_decoder
        ).compound
        if mcstructure.get_int("format_version").py_int == 1:
            min_point = numpy.array(
                tuple(c.py_int for c in mcstructure.get_list("structure_world_origin"))
            )
            max_point = min_point + tuple(
                c.py_int for c in mcstructure.get_list("size")
            )
            selection = SelectionBox(min_point, max_point)
            self._bounds[self.dimensions[0]] = SelectionGroup(selection)
            translator_version = self.translation_manager.get_version(
                "bedrock", (999, 999, 999)
            )
            self._platform = translator_version.platform
            self._version = translator_version.version_number
            structure = mcstructure.get_compound("structure")
            indices = structure.get_list("block_indices")
            blocks_array: numpy.ndarray = numpy.array(
                [[b.py_int for b in layer] for layer in indices],
                dtype=numpy.int32,
            ).reshape((len(indices), *selection.shape))

            palette_tag = structure.get_compound("palette")
            palette_key = next(iter(palette_tag))
            sub_palette_tag = palette_tag.get_compound(palette_key)
            block_palette = list(sub_palette_tag.get_list("block_palette"))

            if -1 in blocks_array[0]:
                blocks_array[0][blocks_array[0] == -1] = len(block_palette)
                block_palette.append(
                    CompoundTag(
                        {
                            "name": StringTag("minecraft:structure_void"),
                            "states": CompoundTag(),
                            "version": IntTag(17694723),  # 1.13.0
                        }
                    )
                )

            for cx, cz in selection.chunk_locations():
                chunk_box = SelectionBox.create_chunk_box(cx, cz).intersection(
                    selection
                )
                array_slice = (slice(None),) + chunk_box.create_moved_box(
                    selection.min, subtract=True
                ).slice
                chunk_blocks_: numpy.ndarray = blocks_array[array_slice]
                chunk_palette_indexes, chunk_blocks = numpy.unique(
                    chunk_blocks_.reshape((chunk_blocks_.shape[0], -1)).T,
                    return_inverse=True,
                    axis=0,
                )
                chunk_blocks = chunk_blocks.reshape(chunk_blocks_.shape[1:])

                chunk_palette = numpy.empty(len(chunk_palette_indexes), dtype=object)
                for palette_index, indexes in enumerate(chunk_palette_indexes):
                    chunk_palette[palette_index] = tuple(
                        block_palette[index] for index in indexes if index >= 0
                    )

                self._chunks[(cx, cz)] = (
                    chunk_box,
                    chunk_blocks,
                    chunk_palette,
                    [],
                    [],
                )

            block_entities = {
                int(key): val["block_entity_data"]
                for key, val in sub_palette_tag.get_compound(
                    "block_position_data"
                ).items()
                if "block_entity_data" in val
            }
            for location, block_entity in block_entities.items():
                if all(key in block_entity for key in "xyz"):
                    x, y, z = (
                        block_entity["x"].py_int,
                        block_entity["y"].py_int,
                        block_entity["z"].py_int,
                    )
                    cx, cz = x >> 4, z >> 4
                    if (cx, cz) in self._chunks and (x, y, z) in self._chunks[(cx, cz)][
                        0
                    ]:
                        self._chunks[(cx, cz)][3].append(block_entity)

            for entity in structure.get_list("entities"):
                if "Pos" in entity:
                    pos = entity.get_list("Pos")
                    x = pos[0].py_float
                    y = pos[1].py_float
                    z = pos[2].py_float
                    cx, cz = numpy.floor([x, z]).astype(int) >> 4
                    if (cx, cz) in self._chunks and (x, y, z) in self._chunks[(cx, cz)][
                        0
                    ]:
                        self._chunks[(cx, cz)][4].append(entity)

        else:
            raise Exception(
                f"mcstructure file with format_version=={mcstructure.get_int('format_version').py_int} cannot be read"
            )

    @staticmethod
    def is_valid(path: str) -> bool:
        return os.path.isfile(path) and path.endswith(".mcstructure")

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {"bedrock": (False, True)}

    @property
    def extensions(self) -> Tuple[str, ...]:
        return (".mcstructure",)

    def _get_interface(self, raw_chunk_data=None) -> "MCStructureInterface":
        return mcstructure_interface

    def _get_interface_and_translator(
        self, raw_chunk_data=None
    ) -> Tuple["Interface", "Translator", VersionNumberAny]:
        interface = self._get_interface(raw_chunk_data)
        translator, version_identifier = interface.get_translator(
            self.max_world_version, raw_chunk_data, self.translation_manager
        )
        return interface, translator, version_identifier

    def save_to(self, f: BinaryIO):
        selection = self._bounds[self.dimensions[0]].selection_boxes[0]
        mcstructure = CompoundTag(
            {
                "format_version": IntTag(1),
                "structure_world_origin": ListTag(
                    [
                        IntTag(selection.min_x),
                        IntTag(selection.min_y),
                        IntTag(selection.min_z),
                    ]
                ),
                "size": ListTag(
                    [
                        IntTag(selection.size_x),
                        IntTag(selection.size_y),
                        IntTag(selection.size_z),
                    ]
                ),
            }
        )

        entities = []
        block_entities = []
        blocks = numpy.zeros(selection.shape, dtype=numpy.uint32)
        palette: List[AnyNDArray] = []
        if self.version < (1, 13, 0):
            raise Exception(
                "Writing to mcstructre files in pre-1.13 format is not currently supported."
            )
        else:
            arr = numpy.empty(1, dtype=object)
            arr[0] = [
                CompoundTag(
                    {
                        "name": StringTag("minecraft:air"),
                        "states": CompoundTag(),
                        "version": IntTag(17694723),
                    }
                )
            ]
            palette.append(arr)

        palette_len = 1

        for (
            selection_,
            blocks_,
            palette_,
            block_entities_,
            entities_,
        ) in self._chunks.values():
            if selection_.intersects(selection):
                box = selection_.create_moved_box(selection.min, subtract=True)
                blocks[box.slice] = blocks_ + palette_len
                palette.append(palette_)
                palette_len += len(palette_)
                block_entities += block_entities_
                entities += entities_

        compact_palette, lut = brute_sort_objects_no_hash(numpy.concatenate(palette))
        blocks = lut[blocks].ravel()
        block_palette = []
        block_palette_indices = []
        for block_list in compact_palette:
            indexed_block = [-1] * 2
            for block_layer, block in enumerate(block_list):
                if block_layer >= 2:
                    break
                if block["name"] != StringTag("minecraft:structure_void"):
                    if block in block_palette:
                        indexed_block[block_layer] = block_palette.index(block)
                    else:
                        indexed_block[block_layer] = len(block_palette)
                        block_palette.append(block)
            block_palette_indices.append(indexed_block)

        block_indices = numpy.array(block_palette_indices, dtype=numpy.int32)[blocks].T

        mcstructure["structure"] = CompoundTag(
            {
                "block_indices": ListTag(
                    [  # a list of tag ints that index into the block_palette. One list per block layer
                        ListTag([IntTag(block) for block in layer])
                        for layer in block_indices
                    ]
                ),
                "entities": ListTag(entities),
                "palette": CompoundTag(
                    {
                        "default": CompoundTag(
                            {
                                "block_palette": ListTag(block_palette),
                                "block_position_data": CompoundTag(
                                    {
                                        str(
                                            (
                                                (
                                                    block_entity["x"].py_int
                                                    - selection.min_x
                                                )
                                                * selection.size_y
                                                + (
                                                    block_entity["y"].py_int
                                                    - selection.min_y
                                                )
                                            )
                                            * selection.size_z
                                            + block_entity["z"].py_int
                                            - selection.min_z
                                        ): CompoundTag(
                                            {"block_entity_data": block_entity}
                                        )
                                        for block_entity in block_entities
                                    }
                                ),
                            }
                        )
                    }
                ),
            }
        )
        mcstructure.save_to(
            f, compressed=False, little_endian=True, string_encoder=utf8_escape_encoder
        )

    def _close(self):
        """Close the disk database"""
        self._chunks.clear()

    def unload(self):
        pass

    def all_chunk_coords(
        self, dimension: Optional[Dimension] = None
    ) -> Iterable[ChunkCoordinates]:
        yield from self._chunks.keys()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return (cx, cz) in self._chunks

    def _encode(
        self,
        interface: MCStructureInterface,
        chunk: Chunk,
        dimension: Dimension,
        chunk_palette: AnyNDArray,
    ):
        return interface.encode(
            chunk,
            chunk_palette,
            self.max_world_version,
            SelectionBox.create_chunk_box(chunk.cx, chunk.cz).intersection(
                self._bounds[dimension].to_box()
            ),
        )

    def _delete_chunk(self, cx: int, cz: int, dimension: Optional[Dimension] = None):
        if (cx, cz) in self._chunks:
            del self._chunks[(cx, cz)]

    def _put_raw_chunk_data(
        self,
        cx: int,
        cz: int,
        section: MCStructureChunk,
        dimension: Optional[Dimension] = None,
    ):
        self._chunks[(cx, cz)] = copy.deepcopy(
            (
                section.selection,
                section.blocks,
                section.palette,
                section.block_entities,
                section.entities,
            )
        )

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: Optional[Dimension] = None
    ) -> MCStructureChunk:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        if (cx, cz) in self._chunks:
            return MCStructureChunk(*copy.deepcopy(self._chunks[(cx, cz)]))
        else:
            raise ChunkDoesNotExist



from abc import abstractmethod
from typing import BinaryIO, List, Any, Tuple, Iterable, Union, Optional, Dict
import os

from .format_wrapper import FormatWrapper, VersionNumberT
from amulet.api.data_types import Dimension
from amulet.api.errors import ObjectReadError, ObjectReadWriteError, PlayerDoesNotExist
from amulet.api.player import Player
from amulet.api.selection import SelectionGroup


class StructureFormatWrapper(FormatWrapper[VersionNumberT]):
    """A base FormatWrapper for all structures that only have one dimension."""

    def __init__(self, path: str):
        """
        Construct a new instance of :class:`StructureFormatWrapper`.

        This should not be used directly. You should instead use :func:`amulet.load_format`.

        :param path: The file path to the serialised data.
        """
        super().__init__(path)
        if self.extensions:
            ext = os.path.splitext(self._path)[1]
            if ext not in self.extensions:
                raise ObjectReadWriteError(
                    f"The given file does not have a valid file extension. Must be one of {self.extensions}"
                )

    @property
    def level_name(self) -> str:
        return os.path.basename(self.path)

    @property
    def dimensions(self) -> List[Dimension]:
        return ["main"]

    @property
    def can_add_dimension(self) -> bool:
        return False

    def register_dimension(self, dimension_identifier: Any):
        pass

    @property
    def requires_selection(self) -> bool:
        return True

    @property
    @abstractmethod
    def extensions(self) -> Tuple[str, ...]:
        """
        The extension the file can have to be valid.

        eg (".schematic",)
        """
        raise NotImplementedError

    def _set_selection(
        self,
        bounds: Union[SelectionGroup, Dict[Dimension, Optional[SelectionGroup]], None],
    ):
        if isinstance(bounds, SelectionGroup):
            bounds = self._clean_selection(bounds)
            self._bounds = {dim: bounds for dim in self.dimensions}
        elif isinstance(bounds, dict):
            for dim in self.dimensions:
                group = bounds.get(dim, None)
                if isinstance(group, SelectionGroup):
                    self._bounds[dim] = self._clean_selection(group)
        else:
            raise ObjectReadError("A selection was required but none were given.")

    @abstractmethod
    def open_from(self, f: BinaryIO):
        """
        Load from a file like object. Useful to load from RAM rather than disk.

        :param f: The file like object to read data from.
        """
        raise NotImplementedError

    def _open(self):
        if not os.path.isfile(self._path):
            raise ObjectReadError(f"There is no file to read at {self._path}")
        with open(self._path, "rb") as f:
            self.open_from(f)
        self._is_open = True
        self._has_lock = True

    @abstractmethod
    def save_to(self, f: BinaryIO):
        """
        Write to a file like object. Useful to write to RAM rather than disk.

        :param f: The file like object to write data to.
        """
        raise NotImplementedError

    def _save(self):
        os.makedirs(os.path.dirname(self._path), exist_ok=True)
        with open(self._path, "wb") as f:
            self.save_to(f)

    def all_player_ids(self) -> Iterable[str]:
        yield from ()

    def has_player(self, player_id: str) -> bool:
        return False

    def _load_player(self, player_id: str) -> Player:
        raise PlayerDoesNotExist

    def _get_raw_player_data(self, player_id: str) -> Any:
        raise PlayerDoesNotExist
